# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Optional

from iotas.note import Note, DirtyFields


class BackupStorage(ABC):
    """Backup storage interface."""

    @abstractmethod
    def add_note(self, note: Note) -> None:
        """Add a new note to the database.

        :param Note note: Note to add
        """
        raise NotImplementedError()

    @abstractmethod
    def get_all_notes_count(self, include_locally_deleted: bool = True) -> int:
        """Fetch the number of notes in the database.

        :param bool include_locally_deleted: Whether to include locally deleted notes in the count
        :return: The count
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def get_all_notes(self, load_content: bool = False) -> list[Note]:
        """Fetch all notes from the database.

        :param bool load_content: Whether to load the content for each note
        :return: List of notes
        :rtype: list[Note]
        """
        raise NotImplementedError()

    @abstractmethod
    def get_note_by_remote_id(self, remote_id: int) -> Optional[Note]:
        """Fetch note from the database by remote id.

        :param int remote_id: Remote id of the note
        :return: The note, or None
        :rtype: Optional[Note]
        """
        raise NotImplementedError()

    @abstractmethod
    def get_note_by_title(self, title: str) -> Optional[Note]:
        """Fetch note from the database by title.

        :param str title: Search title
        :return: The note, or None
        :rtype: Optional[Note]
        """
        raise NotImplementedError()

    @abstractmethod
    def persist_note_selective(self, note: Note, updated_fields: DirtyFields) -> None:
        """Persist local note based on remote changes.

        :param Note note: The note to update
        :param DirtyFields updated_fields: Which fields to update
        """
        raise NotImplementedError()

    @abstractmethod
    def create_duplicate_note(self, note: Note, reason: str) -> Note:
        """Create a duplicate note with a prefixed title.

        :param Note note: The note to duplicate
        :param str reason: The reason for the duplication, which is prefixed on the title
        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gettext import ngettext
from gi.repository import Adw, Gdk, GObject, Gio, GLib, Gtk

import logging
from typing import Callable, Optional

from iotas.category import Category, CategorySpecialPurpose
from iotas.category_manager import CategoryManager
from iotas.config_manager import ConfigManager
import iotas.const as const
from iotas.first_start_page import FirstStartPage
from iotas.index_menu_button import IndexMenuButton
from iotas.index_note_list import IndexNoteList, FocusedRow
from iotas.index_search_header_bar import IndexSearchHeaderBar, IndexSearchState
from iotas.note import Note
from iotas.note_manager import NoteManager
from iotas.selection_header_bar import SelectionHeaderBar
from iotas.sidebar import Sidebar
from iotas.sync_manager import SyncManager
from iotas.theme_selector import ThemeSelector


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/index.ui")
class Index(Adw.BreakpointBin):
    __gtype_name__ = "Index"

    __gsignals__ = {
        # Parameters are the note and whether to open immediately (disabling animation)
        "note-opened": (GObject.SignalFlags.RUN_FIRST, None, (Note, bool)),
        "reauthenticate": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "export": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "header-bar-height": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
    }

    UNDO_NOTE_DELETION_DURATION = 5.0
    BUMPED_OUT_OF_EDITOR_NOTIFICATION_DURATION = 5.0
    SYNC_UPDATE_NOTIFICATION_DURATION = 2.0
    NOTES_APP_MISSING_NOTIFICATION_DURATION = 8.0
    BREAKPOINT_WIDTH = 700

    _overlay_view: Adw.OverlaySplitView = Gtk.Template.Child()
    _breakpoint: Adw.Breakpoint = Gtk.Template.Child()

    _stack: Gtk.Stack = Gtk.Template.Child()
    _sidebar: Sidebar = Gtk.Template.Child()
    _note_list_scroll: Gtk.ScrolledWindow = Gtk.Template.Child()
    _note_list: IndexNoteList = Gtk.Template.Child()
    _first_start: FirstStartPage = Gtk.Template.Child()
    _loading: Adw.StatusPage = Gtk.Template.Child()
    _empty: Adw.StatusPage = Gtk.Template.Child()
    _search_info: Adw.StatusPage = Gtk.Template.Child()
    _search_empty: Adw.StatusPage = Gtk.Template.Child()

    _header_stack: Gtk.Stack = Gtk.Template.Child()
    _main_header_bar: Adw.HeaderBar = Gtk.Template.Child()
    _content_sidebar_button: Gtk.Button = Gtk.Template.Child()
    _search_header_bar: IndexSearchHeaderBar = Gtk.Template.Child()
    _selection_header_bar: SelectionHeaderBar = Gtk.Template.Child()
    _title: Gtk.Label = Gtk.Template.Child()

    _toast_overlay: Adw.ToastOverlay = Gtk.Template.Child()

    _menu_button: IndexMenuButton = Gtk.Template.Child()
    _new_button: Gtk.Button = Gtk.Template.Child()

    _misc_banner: Adw.Banner = Gtk.Template.Child()
    _offline_banner: Adw.Banner = Gtk.Template.Child()
    _app_id_change_reauthenticate_banner: Adw.Banner = Gtk.Template.Child()
    _generic_reauthenticate_banner: Adw.Banner = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        self.__note_manager: NoteManager
        self.__category_manager: CategoryManager
        self.__sync_manager: SyncManager
        self.__config_manager = ConfigManager.get_default()

        self.__active = False
        self.__initialised = False
        self.__import_syncing = False
        self.__search_after_init: Optional[str] = None
        self.__searching_before_selection = False
        self.__header_bar_height_sent = False

        self.__add_fav_action: Gio.SimpleAction
        self.__remove_fav_action: Gio.SimpleAction
        self.__cancel_action: Gio.SimpleAction
        self.__select_all_action: Gio.SimpleAction

        self.__syncing_toast: Optional[Adw.Toast] = None
        self.__focused_row_pre_changes: Optional[FocusedRow] = None
        self.__open_note_row: Optional[FocusedRow] = None

        self.__search_model = None

        # Startup bootstrap category
        self.__category = Category("", 0)
        self.__category.special_purpose = CategorySpecialPurpose.ALL
        self._title.set_label(Category.ALL_TITLE)

        self._note_list.connect(
            "note-opened", lambda _o, note: self.__open_note(note, immediate=False)
        )
        self._note_list.connect("note-checkbox-activated", self.__on_note_checkbox_activated)
        self._note_list.connect(
            "note-checkbox-deactivated",
            lambda _o, note: self._selection_header_bar.set_note_selected(note, False),
        )

        self._selection_header_bar.connect("categories-changed", self.__on_notes_categories_changed)
        self._selection_header_bar.connect("export", self.__on_notes_export)
        self._selection_header_bar.connect("delete", self.__on_notes_delete)
        self._selection_header_bar.connect(
            "set-favourite", lambda _o: self.__update_favourite_on_selection(True)
        )
        self._selection_header_bar.connect(
            "clear-favourite", lambda _o: self.__update_favourite_on_selection(False)
        )
        self._selection_header_bar.connect("abort", lambda _o: self.__cancel_selection())

        self._menu_button.get_popover().add_child(ThemeSelector(), "theme")

        self.connect("realize", self.__on_realize)

        controller = Gtk.EventControllerKey()
        controller.connect("key-pressed", self.__on_index_key_pressed)
        self.add_controller(controller)

        self._search_header_bar.connect(
            "changed",
            lambda _o: self.__apply_search(focused_row=None, invalidate_scroll_and_expansion=True),
        )

        self.__config_manager.connect_changed(
            ConfigManager.INDEX_CATEGORY_STYLE,
            lambda: self._note_list.update_category_labels(self.__category),
        )

        self.__config_manager.connect_changed(
            ConfigManager.PERSIST_SIDEBAR,
            self.__update_for_sidebar_pin_change,
        )

        self._breakpoint.connect("apply", self.__on_breakpoint_apply)
        self._breakpoint.connect("unapply", self.__on_breakpoint_unapply)

        self._overlay_view.connect(
            "notify::show-sidebar", lambda _o, _v: self.__update_header_buttons_for_state()
        )

    def do_size_allocate(self, width: int, height: int, baseline: int) -> None:
        """Allocates widget with a transformation that translates the origin to the position in
        allocation.

        :param int width: Width of the allocation
        :param int height: Height of the allocation
        :param int baseline: The baseline of the child
        """
        # Using the index headerbar stack to get a rough header bar height for the editor,
        # accommodating for text size preferences.
        Adw.BreakpointBin.do_size_allocate(self, width, height, baseline)
        height = self._header_stack.get_height()
        if height and not self.__header_bar_height_sent:
            self.__header_bar_height_sent = True
            self.emit("header-bar-height", height)

    def setup(
        self,
        note_manager: NoteManager,
        category_manager: CategoryManager,
        sync_manager: SyncManager,
    ) -> None:
        """Perform initial setup.

        :param NoteManager note_manager: Note manager
        :param CategoryManager category_manager: Category manager
        :param SyncManager sync_manager: Remote sync manager
        """
        self.__sync_manager = sync_manager

        self.__note_manager = note_manager
        self.__note_manager.connect(
            "initial-load-complete", lambda _o: self.__on_initial_load_from_db_complete()
        )
        self.__note_manager.connect("new-note-persisted", self.__on_new_note_persisted)

        self.__category_manager = category_manager

        self.__setup_actions()

        self.__sync_manager.set_managers(self.__note_manager, self.__category_manager)

        self._sidebar.setup(
            self.__sync_manager,
            self.__category_manager,
            self.__on_category_activated,
        )
        self._search_header_bar.setup(self._note_list, self.__note_manager)
        self._selection_header_bar.setup(self.__category_manager.tree_store)
        self._note_list.setup(self.__note_manager, self.__on_listbox_key_pressed)
        self._selection_header_bar.bind_property(
            "active", self._note_list, "selecting", GObject.BindingFlags.SYNC_CREATE
        )
        self._search_header_bar.bind_property(
            "active", self._note_list, "searching", GObject.BindingFlags.SYNC_CREATE
        )

        if self.__config_manager.first_start:
            self._stack.set_visible_child(self._first_start)
            self._first_start.initialise_device()
        else:
            self._stack.set_visible_child(self._loading)

        self.__sync_manager.connect("ready", self.__on_sync_ready)
        self.__sync_manager.connect("started", self.__on_sync_started)
        self.__sync_manager.connect("have-changes", self.__on_sync_have_changes)
        self.__sync_manager.connect("finished", self.__on_sync_finished)
        self.__sync_manager.connect("missing-password", self.__on_missing_password)
        self.__sync_manager.connect(
            "notes-capability-missing", self.__on_missing_server_notes_capability
        )

        self.__sync_manager.init_auth()

        self.__sync_manager.bind_property(
            "offline", self._offline_banner, "revealed", GObject.BindingFlags.SYNC_CREATE
        )

    def update_for_dialog_visibility(self, visible: bool) -> None:
        """Update if dialogs visible, disabling actions.

        :param bool visible: Whether a dialog is visible
        """
        if self.active:
            for action in (self.__cancel_action, self.__select_all_action):
                action.set_enabled(not visible)

    def update_layout_for_initial_size(self, width: int, height: int) -> None:
        """Update layout for initial window size.

        :param int width: Window width
        :param int height: Window height
        """
        self.__update_for_sidebar_pin_change(width)
        self._note_list.update_search_pagesize_for_height(height)

    def update_search_pagesize_for_height(self, height: int) -> None:
        """Recalculate number of results to show on first search page.

        :param int height: The new window height
        """
        self._note_list.update_search_pagesize_for_height(height)

    def show_note_conflict_alert(self) -> None:
        """Let the user know a sync conflict has occurred with the note being edited."""
        # Translators: Description, notification
        toast = Adw.Toast.new(_("Sync conflict with note being edited"))
        toast.set_timeout(self.BUMPED_OUT_OF_EDITOR_NOTIFICATION_DURATION)
        self._toast_overlay.add_toast(toast)

    def show_editing_note_deleted_alert(self) -> None:
        """Let the user know the note being edited was remotely deleted."""
        # Translators: Description, notification
        toast = Adw.Toast.new(_("The note being edited was remotely deleted"))
        toast.set_timeout(self.BUMPED_OUT_OF_EDITOR_NOTIFICATION_DURATION)
        self._toast_overlay.add_toast(toast)

    def show_secret_service_failure_alert(self) -> None:
        text = _(
            # Translators: Description, notification. "Secret Service" and "gnome-keyring" should
            # likely not be translated.
            "Failure accessing Secret Service. Ensure you have a provider like gnome-keyring which "
            "has a default keyring setup that is unlocked."
        )
        self._misc_banner.set_title(text)
        self._misc_banner.set_revealed(True)

    def refresh_after_note_closed(self, note: Note) -> None:
        """Refresh index after note has been closed.

        :param Note note: The closed note
        """
        if self.__searching():
            self.__apply_search()
        else:
            if self.__open_note_row:
                self.__focused_row_pre_changes = self.__open_note_row
                self.__open_note_row = None
            self.__invalidate_view(check_older_notes=False)
            if not self.__handle_emptied_category(note):
                self._note_list.update_category_labels(self.__category)

        if self.get_root().using_keyboard_navigation:
            if not note.locally_deleted:
                self._note_list.refocus_selected_row()
        else:
            self._note_list.clear_selections()

    def search_from_cli(self, search_term: str) -> None:
        """Initiate a search, from the CLI.

        :param str search_term: The search term
        """
        if self.__initialised:
            self.__start_search_from_cli(search_term)
        else:
            self.__search_after_init = search_term

    def delete_notes(self, notes: list[Note]) -> None:
        """Delete the specified notes, flushing any undo-pending notes first.

        :param list[Note] notes: Notes to delete
        """
        focused_row = self._note_list.get_current_selection()
        if self.__selecting():
            self.__cancel_selection()
        self.__sync_manager.flush_pending_deletions()
        self.__note_manager.delete_notes(notes, provide_undo=True)
        self.__update_for_note_deletions(notes, focused_row)

    def get_selection(self) -> list[Note]:
        """Get selected note/s.

        :return: Selected notes
        :rtype: list[Note]
        """
        if self._selection_header_bar.active:
            return self._selection_header_bar.get_selected()
        else:
            note = self._note_list.get_selected_note()
            if note:
                return [note]
            else:
                return []

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value
        self.__enable_actions(value)

    @Gtk.Template.Callback()
    def _on_reauthenticate(self, _button: Gtk.Button) -> None:
        self.emit("reauthenticate")

    @Gtk.Template.Callback()
    def _on_misc_banner_button(self, _button: Gtk.Button) -> None:
        if self._misc_banner.get_revealed():
            self._misc_banner.set_revealed(False)

    @Gtk.Template.Callback()
    def _on_edge_overshot(self, _obj: Gtk.ScrolledWindow, pos: Gtk.PositionType) -> None:
        if pos == Gtk.PositionType.BOTTOM:
            if self.__searching():
                if self._note_list.have_more_search_results:
                    self._note_list.show_remaining_search_results()
            else:
                if not self._note_list.older_notes_displayed:
                    self.__load_older_notes()

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("toggle-sidebar")
        action.connect("activate", lambda _a, _p: self.__toggle_sidebar())
        action_group.add_action(action)
        app.set_accels_for_action("index.toggle-sidebar", ["F9"])

        action = Gio.SimpleAction.new("category-context-dependent")
        action.connect("activate", lambda _a, _p: self.__on_category_shortcut())
        action_group.add_action(action)
        app.set_accels_for_action("index.category-context-dependent", ["<Control>e"])

        action = Gio.SimpleAction.new("create-note")
        action.connect("activate", lambda _o, _v: self.__create_and_open_note())
        action_group.add_action(action)
        app.set_accels_for_action("index.create-note", ["<Control>n"])

        action = Gio.SimpleAction.new("create-note-with-content", GLib.VariantType("s"))
        action.connect("activate", self.__on_create_note_with_content)
        action_group.add_action(action)

        action = Gio.SimpleAction.new("export-notes")
        action.connect("activate", lambda _a, _p: self.emit("export"))
        action_group.add_action(action)
        app.set_accels_for_action("index.export-notes", ["<Shift><Control>s"])

        action = Gio.SimpleAction.new("delete-from-index")
        action.connect("activate", lambda _a, _p: self.__delete_note_from_keyboard())
        action_group.add_action(action)
        app.set_accels_for_action("index.delete-from-index", ["<Shift>Delete"])

        action = Gio.SimpleAction.new("load-older-notes")
        action.connect("activate", lambda _a, _p: self.__load_older_notes())
        action_group.add_action(action)

        action = Gio.SimpleAction.new("show-more")
        action.connect("activate", lambda _a, _p: self._note_list.show_remaining_search_results())
        action_group.add_action(action)

        action = Gio.SimpleAction.new("enter-search")
        action.connect("activate", lambda _a, _p: self.__enter_search(clear_text=True))
        action_group.add_action(action)
        app.set_accels_for_action("index.enter-search", ["<Control>f"])

        action = Gio.SimpleAction.new("select-notes")
        action.connect("activate", lambda _a, _p: self.__start_selection())
        action_group.add_action(action)
        app.set_accels_for_action("index.select-notes", ["<Control>s"])

        action = Gio.SimpleAction.new("cancel")
        action.connect("activate", lambda _a, _p: self.__cancel_out_of_context(False))
        action_group.add_action(action)
        self.__cancel_action = action
        app.set_accels_for_action("index.cancel", ["Escape", "<Alt>Left"])

        action = Gio.SimpleAction.new("reset-filters")
        action.connect("activate", lambda _a, _p: self.__cancel_out_of_context(True))
        action_group.add_action(action)
        app.set_accels_for_action(
            "index.reset-filters", ["<Control><Shift>BackSpace", "<Control><Shift>Delete"]
        )

        action = Gio.SimpleAction.new("selection-toggle-favourite")
        action.connect("activate", lambda _a, _p: self.__on_toggle_favourite())
        action_group.add_action(action)
        app.set_accels_for_action("index.selection-toggle-favourite", ["<Control>g"])

        action = Gio.SimpleAction.new("show-menu")
        action.connect("activate", lambda _a, _p: self._menu_button.popup())
        action_group.add_action(action)
        app.set_accels_for_action("index.show-menu", ["F10"])

        action = Gio.SimpleAction.new("select-all")
        action.connect("activate", lambda _a, _p: self.__select_all())
        action_group.add_action(action)
        self.__select_all_action = action
        app.set_accels_for_action("index.select-all", ["<Control>a"])

        self.__action_group = action_group
        app.get_active_window().insert_action_group("index", action_group)

    def __enable_actions(self, enabled: bool) -> None:
        """Toggle whether index actions are enabled.

        :param bool enabled: New value
        """
        actions = self.__action_group.list_actions()
        for action in actions:
            self.__action_group.lookup_action(action).set_enabled(enabled)

    def __on_realize(self, _widget: Gtk.Widget) -> None:
        # Preset filters to avoid momentarily populating all at startup
        self.__note_manager.update_filters(
            self.__category, self._note_list.older_notes_displayed, check_older_notes=False
        )
        self.__note_manager.initiate_model_from_db()
        self.__category_manager.populate()

    def __on_create_note_with_content(
        self,
        _obj: GObject.Object,
        param: GObject.ParamSpec,
    ) -> None:
        """Create a new note with content and edit it."""
        self.__create_and_open_note(param.get_string())

    def __on_notes_categories_changed(self, _obj: SelectionHeaderBar) -> None:
        focused_row = self._note_list.get_current_selection()
        changeset = self._selection_header_bar.get_categories_changeset()
        self.__cancel_selection()
        for note, old_category in changeset:
            self.__note_manager.persist_note_category(note, old_category)
        if self.__searching():
            self.__apply_search(focused_row)
        else:
            self.__invalidate_view(check_older_notes=False)
            self.__handle_emptied_category()
            if focused_row:
                self._note_list.place_focus_after_change(focused_row)

    def __on_notes_export(self, _obj: SelectionHeaderBar) -> None:
        self.emit("export")

    def __on_notes_delete(self, _obj: SelectionHeaderBar) -> None:
        notes = self._selection_header_bar.get_selected()
        self.delete_notes(notes)

    def __on_note_checkbox_activated(self, _obj: GObject.Object, note: Note) -> None:
        if not self.__selecting():
            self.__start_selection()
        self._selection_header_bar.set_note_selected(note, True)

    def __on_category_activated(self, category: Category, close_sidebar: bool = True) -> None:
        self.__category = category

        self._note_list.disconnect_older_notes_section()

        # Handle clicking on category while searching or selecting with a persistent sidebar
        if self.__searching():
            self.__end_search()
            # return as __end_search itself calls __on_category_activated
            return
        elif self.__selecting():
            self.__searching_before_selection = False
            self.__cancel_selection()

        self._title.set_label(category.display_name)
        self._note_list_scroll.get_vadjustment().set_value(0)
        self.__invalidate_view(check_older_notes=True)

        self._note_list.refresh_section_visibility(self.__category)
        self._note_list.update_category_labels(self.__category)
        if close_sidebar:
            self.__close_sidebar()
        if self.get_root().using_keyboard_navigation:
            self._note_list.move_focus_to_list_top()

    def __on_sync_ready(self, _obj: GObject.Object, new_setup: bool) -> None:
        if new_setup:
            self.__import_syncing = True
            if self.__searching():
                self.__end_search()
        if self._app_id_change_reauthenticate_banner.get_revealed():
            self._app_id_change_reauthenticate_banner.set_revealed(False)
        if self._generic_reauthenticate_banner.get_revealed():
            self._generic_reauthenticate_banner.set_revealed(False)

    def __on_sync_started(self, _obj: GObject.Object) -> None:
        if self.active and self.__config_manager.show_syncing_debug_notification:
            # Another toast misuse replacing a revealer notification. Debug only at least.
            # TODO in future look at replacing this with a (debug only) spinner.

            # Translators: Description, notification
            self.__syncing_toast = Adw.Toast.new(_("Syncing"))

            # A cludge, only passable for this debug functionality.
            self.__syncing_toast.set_timeout(60.0)
            self._toast_overlay.add_toast(self.__syncing_toast)

    def __on_sync_have_changes(self, _obj: GObject.Object) -> None:
        if self.active:
            self.__focused_row_pre_changes = self._note_list.get_current_selection()

    def __on_sync_finished(self, _obj: SyncManager, changes: int) -> None:
        if self.active:
            if self.__syncing_toast:
                self.__syncing_toast.dismiss()
                self.__syncing_toast = None
            if self.__import_syncing:
                self.__on_category_activated(self.__category, close_sidebar=True)
                self.__import_syncing = False
            elif changes > 0:
                self.__invalidate_view(check_older_notes=False)
                self.__focused_row_pre_changes = None
                self._note_list.update_category_labels(self.__category)

                # Translators: Description, notification, {} is a number
                msg = ngettext("{} change", "{} changes", changes).format(changes)
                toast = Adw.Toast.new(msg)
                toast.set_timeout(self.SYNC_UPDATE_NOTIFICATION_DURATION)
                self._toast_overlay.add_toast(toast)

    def __on_missing_password(self, _obj: GObject.Object) -> None:
        app = Gio.Application.get_default()
        previous_version = app.previous_version
        if self.__sync_manager.configured_but_no_password:
            if previous_version != "" and previous_version < const.APP_ID_CHANGE_VERSION:
                self._app_id_change_reauthenticate_banner.set_revealed(True)
            else:
                self._generic_reauthenticate_banner.set_revealed(True)

    def __on_missing_server_notes_capability(self, _obj: GObject.Object) -> None:
        # Translators: Description, notification
        msg = _("Sync failure. Is the Nextcloud Notes app installed on the server?")
        toast = Adw.Toast.new(msg)
        toast.set_timeout(self.NOTES_APP_MISSING_NOTIFICATION_DURATION)
        self._toast_overlay.add_toast(toast)

    def __on_index_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval in (Gdk.KEY_Down, Gdk.KEY_KP_Down, Gdk.KEY_Up, Gdk.KEY_KP_Up):
            self.get_root().using_keyboard_navigation = True
            if self._overlay_view.get_pin_sidebar() or not self._overlay_view.get_show_sidebar():
                if self._note_list.move_focus_to_first_visible(
                    self._note_list_scroll.get_vadjustment().get_value()
                ):
                    return Gdk.EVENT_STOP
            else:
                # If we have an auto-expanded sidebar use the shortcut to send the focus
                self._sidebar.take_focus()
                return Gdk.EVENT_STOP
        elif not self.__searching() and not self.__selecting():
            if self._search_header_bar.check_if_starting(controller, keyval, state):
                self.__enter_search(clear_text=False)
                return Gdk.EVENT_STOP
        elif self.__searching() and not self.__selecting():
            if self._search_header_bar.check_if_starting(controller, keyval, state, clear=False):
                self._search_header_bar.focus_entry()
                return Gdk.EVENT_STOP
            elif self._search_header_bar.check_for_open_first_result_shortcut(
                controller, keyval, state
            ):
                # Check that we have some results, a little indirect
                if self._stack.get_visible_child() == self._note_list_scroll:
                    self._note_list.open_first_search_result()
                    return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def __on_listbox_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval in (Gdk.KEY_Down, Gdk.KEY_KP_Down):
            self.get_root().using_keyboard_navigation = True
            self._note_list.focus_next_list_row(self.__get_focused())
            return Gdk.EVENT_STOP
        elif keyval in (Gdk.KEY_Up, Gdk.KEY_KP_Up):
            self.get_root().using_keyboard_navigation = True
            self._note_list.focus_previous_list_row(self.__get_focused())
            return Gdk.EVENT_STOP
        elif not self.__searching() and not self.__selecting():
            if self._search_header_bar.check_if_starting(controller, keyval, state):
                self.__enter_search(clear_text=False)
                return Gdk.EVENT_STOP
        elif self.__searching() and not self.__selecting():
            if self._search_header_bar.check_if_starting(controller, keyval, state, clear=False):
                self._search_header_bar.focus_entry()
                return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def __on_toggle_favourite(self) -> None:
        if self.__selecting():
            self._selection_header_bar.toggle_favourite_on_selection()
        else:
            note = None
            window = self.get_root()
            focused = window.get_focus()
            if isinstance(focused, Gtk.ListBoxRow):
                note = focused.get_first_child().note
            if note:
                self.__update_favourite_on_notes([note], not note.favourite)

    def __on_category_shortcut(self) -> None:
        if self.__selecting():
            self._selection_header_bar.edit_category_for_selection()
        else:
            self.__toggle_sidebar()

    def __on_breakpoint_apply(self, _bp: Adw.Breakpoint) -> None:
        if self._overlay_view.get_pin_sidebar():
            self._overlay_view.set_pin_sidebar(False)
            self._overlay_view.set_collapsed(True)

        self._overlay_view.set_show_sidebar(False)
        self.__update_header_buttons_for_state()

    def __on_breakpoint_unapply(self, _bp: Adw.Breakpoint) -> None:
        if self.__config_manager.pin_sidebar:
            self._overlay_view.set_pin_sidebar(True)
            self._overlay_view.set_collapsed(False)
            self._overlay_view.set_show_sidebar(True)

        self.__update_header_buttons_for_state()

    def __load_older_notes(self, after: Optional[Callable] = None) -> None:
        # Somewhat clunky misuse of toast to replace previous revealer notification
        # Translators: Description, notification
        toast = Adw.Toast.new(_("Loading"))
        toast.set_timeout(300.0)
        self._toast_overlay.add_toast(toast)

        def load(toast: Adw.Toast, after: Optional[Callable]) -> None:
            if self.__note_manager.older_notes_model is None:
                self.__load_older_notes_worker()
            else:
                self._note_list.populate_older_notes(self.__note_manager.older_notes_model)
            self.__invalidate_view(check_older_notes=False)
            self._note_list.update_category_labels(self.__category)
            toast.dismiss()
            if after is not None:
                after()

        GLib.timeout_add(175, load, toast, after)

    def __create_and_open_note(self, content: Optional[str] = None) -> None:
        note = self.__note_manager.create_note(self.__category)
        # If we're creating a note from a selection persist it straight away
        if content is not None and content != "":
            note.content = content
            note.update_title_from_top_line()
            self.__note_manager.persist_note_while_editing(note, update_excerpt=True)
        self.__open_note(note, immediate=False)

    def __delete_note_from_keyboard(self) -> None:
        if self.__selecting():
            self._selection_header_bar.handle_delete_keyboard_shortcut()
        else:
            note = None
            window = self.get_root()
            focused = window.get_focus()
            if isinstance(focused, Gtk.ListBoxRow):
                note = focused.get_first_child().note
            if note:
                self.delete_notes([note])

    def __update_for_note_deletions(
        self, notes: list[Note], focused_row: Optional[FocusedRow]
    ) -> None:
        """Perform updates for deleted notes.

        :param list[Note] notes: The deleted notes
        :param FocusedRow focused_row: Row that was focused before the deletion, optional
        """
        if len(notes) > 1:
            # Translators: Description, notification, {} is a positive number
            msg = _("{} notes deleted").format(len(notes))
        else:
            # Translators: Description, notification
            msg = _("Note deleted")

        toast = Adw.Toast.new(msg)
        toast.set_timeout(self.UNDO_NOTE_DELETION_DURATION)
        # Translators: Button
        toast.set_button_label(_("Undo"))
        toast.connect("button-clicked", lambda _o: self.__undo_deletion())
        toast.connect("dismissed", lambda _o: self.__note_manager.on_deletion_undo_elapsed())
        self._toast_overlay.add_toast(toast)

        if self.__searching():
            self.__apply_search(focused_row)
        else:
            self.__invalidate_view(check_older_notes=False)
            self.__handle_emptied_category()
            if focused_row:
                self._note_list.place_focus_after_change(focused_row)

    def __invalidate_view(self, check_older_notes: bool) -> None:
        """Refresh the view."""
        # Update display of empty state, list, etc
        if self.__searching():
            self.__apply_search()
        else:
            self.__note_manager.invalidate_sort()
            older_loaded = self.__note_manager.update_filters(
                self.__category, self._note_list.older_notes_displayed, check_older_notes
            )
            if older_loaded:
                self._note_list.populate_older_notes(self.__note_manager.older_notes_model)

            if (
                self.__focused_row_pre_changes
                and len(self._note_list.get_model(self.__focused_row_pre_changes.listbox)) == 0
            ):
                self._note_list.place_focus_after_change(self.__focused_row_pre_changes)

            count = self.__note_manager.get_filtered_note_count(
                self._note_list.older_notes_displayed
            )
            if not self.__config_manager.first_start:
                if count == 0:
                    self._stack.set_visible_child(self._empty)
                elif self._stack.get_visible_child() in (
                    self._empty,
                    self._first_start,
                    self._loading,
                ):
                    self._stack.set_visible_child(self._note_list_scroll)

    def __start_search_from_cli(self, search_term: str) -> None:
        self._search_header_bar.text = search_term
        self.__enter_search(clear_text=False, avoid_select_all=True)

    def __enter_search(self, clear_text, avoid_select_all: bool = False) -> None:
        """Enter the search UI."""
        if self.__searching():
            # Grab focus and select entry text
            self._note_list.clear_selections()
            self._search_header_bar.enter(False, not avoid_select_all)
            return

        self.__close_sidebar()

        self._search_header_bar.enter(clear_text, False)

        all_category = self._sidebar.select_and_fetch_category(None, CategorySpecialPurpose.ALL)
        assert all_category
        self.__category = all_category

        self.__set_search_visible(True)

        if self.__note_manager.search_model is None:
            GLib.idle_add(self.__init_search_model)

    def __init_search_model(self) -> None:
        model = self.__note_manager.initiate_search_model()
        self._note_list.populate_search(model)

    def __apply_search(
        self,
        focused_row: Optional[FocusedRow] = None,
        invalidate_scroll_and_expansion: bool = False,
    ) -> None:
        if not self.__searching():
            return

        # If trying to search without caching completed delay the search
        if self.__note_manager.search_model is None:
            GLib.timeout_add(500, self.__apply_search, focused_row, invalidate_scroll_and_expansion)
            logging.debug("Delaying search due to model not ready")
            return

        show_all = (
            not invalidate_scroll_and_expansion and self._note_list.showing_expanded_search_set
        )
        result = self._search_header_bar.search(show_all)
        if result == IndexSearchState.RESULTS:
            self._stack.set_visible_child(self._note_list_scroll)
            if invalidate_scroll_and_expansion:
                self._note_list_scroll.get_vadjustment().set_value(0)
            elif focused_row:
                self._note_list.place_focus_after_change(focused_row)
        elif result == IndexSearchState.EMPTY:
            self._stack.set_visible_child(self._search_empty)
        elif result == IndexSearchState.NO_TERM:
            self._stack.set_visible_child(self._search_info)
            self._note_list.restrict_for_search_by_ids(None)

    def __start_selection(self) -> None:
        self.__close_sidebar()
        self.__searching_before_selection = self.__searching()
        self._header_stack.set_visible_child(self._selection_header_bar)
        self._selection_header_bar.activate()
        self.__update_enabled_actions(False)

    def __select_all(self) -> None:
        if self.__searching() and self._search_header_bar.entry_focused:
            self._search_header_bar.select_all_text()
            return
        if self.__selecting() and self._selection_header_bar.category_entry_focused:
            self._selection_header_bar.select_all_category_text()
            return

        if self.__searching():
            if self._note_list.have_more_search_results:
                self._note_list.show_remaining_search_results()
        else:
            if not self._note_list.older_notes_displayed:
                self.__load_older_notes(self.__select_all)
                return

        if not self.__selecting():
            self.__start_selection()
        for note in self._note_list.iter_visible_notes():
            self._selection_header_bar.set_note_selected(note, True)

    def __cancel_selection(self) -> None:
        if self.__searching_before_selection:
            self._header_stack.set_visible_child(self._search_header_bar)
        else:
            self._header_stack.set_visible_child(self._main_header_bar)
        self._selection_header_bar.active = False
        self._note_list.clear_all_checkboxes()
        self.__update_enabled_actions(True)

        if self.get_root().using_keyboard_navigation:
            self._note_list.refocus_selected_row()
        else:
            self._note_list.clear_selections()

    def __searching(self) -> bool:
        return self._search_header_bar.active

    def __selecting(self) -> bool:
        return self._selection_header_bar.active

    def __set_search_visible(self, visible: bool) -> None:
        self._note_list.refresh_section_visibility(self.__category)
        if visible:
            self._header_stack.set_visible_child(self._search_header_bar)
            self._stack.set_visible_child(self._search_info)
        else:
            self._header_stack.set_visible_child(self._main_header_bar)
            self._stack.set_visible_child(self._note_list_scroll)
            if self.__note_manager.search_model is not None:
                self._note_list.restrict_for_search_by_ids(None)

    def __end_search(self) -> None:
        self._search_header_bar.exit()
        note = self._note_list.get_selected_note()
        self._note_list_scroll.get_vadjustment().set_value(0)
        self.__set_search_visible(False)
        # Could look at removing the category updating and filter invalidating parts of this call
        self.__on_category_activated(self.__category, close_sidebar=True)
        if note is not None:
            self._note_list.select_and_focus_note(note)
        else:
            self._note_list.move_focus_to_list_top()

    def __update_enabled_actions(self, enable: bool) -> None:
        for action_name in ("toggle-sidebar", "create-note", "enter-search", "select-notes"):
            action = self.__action_group.lookup(action_name)
            action.set_property("enabled", enable)

    def __load_older_notes_worker(self) -> None:
        model = self.__note_manager.older_notes_model
        if model is None:
            model = self.__note_manager.initiate_older_notes_model()
        self.__note_manager.filter_older_notes_by_date()

        loaded_notes = self._note_list.populate_older_notes(model)
        if loaded_notes:
            if self._stack.get_visible_child() in (self._empty, self._first_start):
                self._stack.set_visible_child(self._note_list_scroll)

    def __on_initial_load_from_db_complete(self) -> None:
        self.__invalidate_view(check_older_notes=True)
        self._note_list.refresh_section_visibility(self.__category)
        self._note_list.clear_selections()

        self.__initialised = True
        GLib.idle_add(self._note_list.update_category_labels, self.__category)
        if self.__search_after_init is not None:
            GLib.idle_add(self.__start_search_from_cli, self.__search_after_init)

    def __on_new_note_persisted(self, _obj: GObject.Object) -> None:
        self.__invalidate_view(check_older_notes=False)
        self._note_list.update_category_labels(self.__category)

    def __open_note(self, note: Note, immediate: bool) -> None:
        self.__open_note_row = self._note_list.get_current_selection()
        self.emit("note-opened", note, immediate)

    def __undo_deletion(self) -> None:
        if self.__note_manager.undo_deletion():
            self.__invalidate_view(check_older_notes=False)

    def __get_focused(self) -> Optional[Gtk.Widget]:
        window = self.get_root()
        return window.get_focus()

    def __handle_emptied_category(self, note: Optional[Note] = None) -> bool:
        # If we're not filtered by category this won't be an issue
        if not self.__is_showing_category():
            return False

        # Including "older notes" here as we always show older notes when displaying a category
        view_empty = self.__note_manager.get_filtered_note_count(include_older_notes=True) == 0
        if view_empty:
            if note is not None and note.category != "":
                new_category = self._sidebar.select_and_fetch_category(note.category)
            else:
                new_category = self._sidebar.select_and_fetch_category(
                    None, CategorySpecialPurpose.ALL
                )
            # Largely form mypy
            if not new_category:
                logging.warning("Failed to find category?")
                return False
            self.__on_category_activated(new_category, close_sidebar=True)
            return True

        return False

    def __update_favourite_on_selection(self, value: bool) -> None:
        """Change whether selected notes are in favourites.

        :param bool value: The new value
        """
        notes = self._selection_header_bar.get_selected()
        self.__update_favourite_on_notes(notes, value)

    def __update_favourite_on_notes(self, notes: list[Note], value: bool) -> None:
        """Change whether notes are in favourites.

        :param list[Note] notes: Notes to update
        :param bool value: The new value
        """
        focused_row = self._note_list.get_current_selection()
        self.__cancel_selection()
        if self.__note_manager.set_and_persist_favourite_for_notes(notes, value):
            self.__sync_manager.sync_now()
            if self.__searching():
                self.__apply_search()
            else:
                self.__invalidate_view(check_older_notes=False)
                if focused_row:
                    self._note_list.place_focus_after_change(focused_row)

    def __cancel_out_of_context(self, including_category: bool) -> None:
        if including_category:
            if (
                self.__category.special_purpose is None
                or self.__category.special_purpose != CategorySpecialPurpose.ALL
            ):
                category, _ = self.__category_manager.get_special_purpose_category(
                    CategorySpecialPurpose.ALL
                )
                # close_sidebar is False as we're closing it below
                self.__on_category_activated(category, close_sidebar=False)
            else:
                self._note_list.disconnect_older_notes_section()
                self._note_list.refresh_section_visibility(self.__category)
                self._note_list_scroll.get_vadjustment().set_value(0)
        closed_the_sidebar = self.__close_sidebar()
        if not closed_the_sidebar:
            if self.__selecting():
                self.__cancel_selection()
            elif self.__searching():
                self.__end_search()

    def __close_sidebar(self) -> bool:
        if self.__config_manager.pin_sidebar and self.get_width() >= self.BREAKPOINT_WIDTH:
            return False
        elif not self._overlay_view.get_show_sidebar():
            return False

        self._overlay_view.set_show_sidebar(False)
        if self.get_root().using_keyboard_navigation:
            self._note_list.grab_focus()
        self.__update_header_buttons_for_state()
        return True

    def __toggle_sidebar(self) -> None:
        if self.__config_manager.pin_sidebar and self.get_width() >= self.BREAKPOINT_WIDTH:
            # Pinned sidebar is open, let's just throw the focus around (as this block can only be
            # entered via keyboard shortcut)
            self.get_root().using_keyboard_navigation = True
            if self._sidebar.has_focus_within():
                self._note_list.grab_focus()
            else:
                self._sidebar.take_focus()
        else:
            if self._overlay_view.get_show_sidebar():
                self._overlay_view.set_show_sidebar(False)
                if self.get_root().using_keyboard_navigation:
                    self._note_list.grab_focus()
            else:
                self._overlay_view.set_show_sidebar(True)
                self._sidebar.take_focus()

    def __update_for_sidebar_pin_change(self, width: int = -1) -> None:
        if width == -1:
            width = self.get_width()
        if width >= self.BREAKPOINT_WIDTH:
            pin = self.__config_manager.pin_sidebar
            self._overlay_view.set_pin_sidebar(pin)
            self._overlay_view.set_show_sidebar(pin)
            self._overlay_view.set_collapsed(not pin)
        else:
            self._overlay_view.set_pin_sidebar(False)
            self._overlay_view.set_show_sidebar(False)
            self._overlay_view.set_collapsed(True)
        self.__update_header_buttons_for_state(width)

    def __update_header_buttons_for_state(self, width: int = -1) -> None:
        if width == -1:
            width = self.get_width()
            # Used during initialisation
            if width == 0:
                width = self.get_property("width-request")
        if width >= self.BREAKPOINT_WIDTH and self.__config_manager.pin_sidebar:
            show_sidebar_close = False
            show_content_sidebar_button = False
            show_sidebar_menu = True
        else:
            show_sidebar_close = True
            show_content_sidebar_button = True
            show_sidebar_menu = False
        self._content_sidebar_button.set_visible(show_content_sidebar_button)
        self._sidebar.show_buttons(show_sidebar_close, show_sidebar_menu)
        self._menu_button.set_visible(not show_sidebar_menu)

    def __perform_initial_sync(self) -> None:
        self.__initial_sync_delay_complete = True
        self.__sync_manager.sync_now()

    def __is_showing_category(self) -> bool:
        return self.__category.special_purpose is None



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
import warnings

from iotas.attachment import Attachment

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_duplicate(self) -> None:
        attachment = Attachment()
        attachment.note_id = 1
        attachment.note_remote_id = 2
        attachment.path = "image.png"

        dupe = attachment.duplicate()
        self.assertIsNotNone(dupe)
        self.assertEqual(dupe.note_id, 1)
        self.assertEqual(dupe.note_remote_id, 2)
        self.assertEqual(dupe.path, "image.png")

    def test_filename(self) -> None:
        attachment = Attachment()
        attachment.path = "dir/subdir/image.png"
        self.assertEqual(attachment.filename, "image.png")

    def test_path_quoted(self) -> None:
        attachment = Attachment()
        attachment.path = "image (1).png"
        self.assertEqual(attachment.path_quoted, "image%20%281%29.png")



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib

import os
import pytest
import shutil
from typing import Callable
import unittest
import warnings

from test_note_exporter import MockHtmlGenerator

from iotas.markdown_render_view import MarkdownRenderView
from iotas.note import Note
from iotas.webkit_pdf_exporter import WebKitPdfExporter

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    exporter: WebKitPdfExporter
    generator = MockHtmlGenerator
    success = False

    def test_export(self) -> None:
        # TODO Skipped during CI for segfault
        if "DISPLAY" not in os.environ:
            return pytest.skip("DISPLAY not set")

        out_dir = self.__prepare_output_dir()
        out_path = os.path.join(out_dir, "out.pdf")

        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"

        self.__reset()
        self.__call_and_wait(self.exporter.export, note, out_path)
        self.assertTrue(self.success)

        self.assertTrue(os.path.exists(out_path))
        self.assertGreater(os.path.getsize(out_path), 20000)

        # TODO improve validation of content. pdfminer.six below was having an issue building in the
        # flatpak toolchain.
        # from subprocess import run
        # p = run(
        #     [f"pdf2txt.py {out_path}"], shell=True, universal_newlines=True, capture_output=True
        # )
        # if p.returncode != 0:
        #     print(p.stderr)
        # self.assertEqual(p.returncode, 0)
        # self.assertEqual(p.stdout.strip(), self.generator.get_output(note))

        self.__clean_output_dir()

    def __reset(self, tex_support=False) -> None:
        self.generator = MockHtmlGenerator()
        render_view = MarkdownRenderView()
        render_view.setup(self.generator, config_manager=None, track_timing=False)
        self.exporter = WebKitPdfExporter(render_view, keep_webkit_process=False)
        self.exporter.set_callbacks(self.__on_finished, self.__on_failed)

    def __prepare_output_dir(self) -> str:
        out_path = self.__get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)
        os.makedirs(out_path)
        return out_path

    def __clean_output_dir(self) -> None:
        out_path = self.__get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)

    def __get_output_dir(self) -> str:
        file_dir = os.path.dirname(__file__)
        return os.path.join(file_dir, os.pardir, "testing-tmp")

    def __call_and_wait(self, call: Callable, *args) -> None:
        context = GLib.MainContext.default()
        self.success = False
        self.awaiting_signal = True
        call(*args)
        while self.awaiting_signal:
            context.iteration(False)

    def __on_finished(self) -> None:
        self.awaiting_signal = False
        self.success = True

    def __on_failed(self, error) -> None:
        print(f"Failed: {error}")
        self.awaiting_signal = False



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject

import logging
from typing import NamedTuple, Optional
from urllib.parse import urlparse, parse_qs

from iotas.note import Note


class UriParseResult(NamedTuple):

    success: bool
    note: Optional[Note] = None
    open: bool = False


# Note: Iotas' addition of this URI handler is a stopgap measure until the freedesktop
# intent system is fleshed out. The handling provided here will be unceremoniously removed
# when that system is available and may be removed or modified without notice beforehand;
# this is not a stable public interface.


class UriToNoteCreator(GObject.Object):
    """Handles creation of notes from temporary iotas-notes:// URI."""

    def __init__(self) -> None:
        super().__init__()

    def parse(self, uri: str) -> UriParseResult:
        """Parse URI to note.

        See comment in source re. interface.

        :param str uri: URI
        :return: Result
        :rtype: UriParseResult
        """
        logging.debug(f"Handling {uri}")
        parsed = urlparse(uri)
        if parsed.scheme != "iotas-notes":
            logging.warning(f"Not handling schema '{parsed.scheme}'")
            return UriParseResult(False)
        if parsed.netloc != "create-note":
            logging.warning(f"Not handling unknown URI '{parsed.netloc}'")
            return UriParseResult(False)

        fields = parse_qs(parsed.query)
        if "content" not in fields or fields["content"][0].strip() == "":
            logging.warning("No content provided, not creating note from URI")
            return UriParseResult(False)
        if "title" not in fields or fields["title"][0].strip() == "":
            logging.warning("No title provided, not creating note from URI")
            return UriParseResult(False)

        note = Note()
        note.content = fields["content"][0]
        note.title = fields["title"][0]
        if "category" in fields and fields["category"][0].strip() != "":
            note.category = fields["category"][0]
        note.flag_changed()

        open = "open" in fields and fields["open"][0] == "1"

        return UriParseResult(True, note, open)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from gi.repository import Adw, Gio, GLib, GObject

from enum import StrEnum
import logging
from typing import Callable, Optional


class HeaderBarVisibility(StrEnum):
    """Top or bottom bar visibility."""

    ALWAYS_VISIBLE = "always-visible"
    AUTO_HIDE = "auto-hide"
    AUTO_HIDE_FULLSCREEN_ONLY = "auto-hide-fullscreen-only"
    DISABLED = "disabled"


class ConfigManager(GObject.Object):

    STYLE = "style-variant"
    WINDOW_SIZE = "window-size"
    FIRST_START = "first-start"
    FONT_SIZE = "font-size"
    LINE_LENGTH = "line-length"
    USE_MONOSPACE_FONT = "use-monospace-font"
    EDITOR_THEME = "editor-theme"
    EDITOR_FORMATTING_BAR_VISIBILTY = "editor-formatting-bar-visibility"
    EDITOR_HEADER_BAR_VISIBILTY = "editor-header-bar-visibility"
    BACKUP_NOTE_EXTENSION = "backup-note-extension"
    PERSIST_SIDEBAR = "persist-sidebar"
    INDEX_CATEGORY_STYLE = "index-category-style"

    NEXTCLOUD_ENDPOINT = "nextcloud-endpoint"
    NEXTCLOUD_USERNAME = "nextcloud-username"
    NEXTCLOUD_PRUNE_THRESHOLD = "nextcloud-prune-threshold"
    SYNC_INTERVAL = "sync-interval"
    NETWORK_TIMEOUT = "network-timeout"

    SPELLING_ENABLED = "spelling-enabled"
    SPELLING_LANGUAGE = "spelling-language"

    SHOW_STARTUP_SECRET_SERVICE_FAILURE = "show-startup-secret-service-failure"
    SHOW_SYNCING_DEBUG_NOTIFICATION = "show-syncing-debug-notification"

    MARKDOWN_RENDER = "markdown-render-enabled"
    MARKDOWN_DETECT_SYNTAX = "markdown-syntax-detection-enabled"
    MARKDOWN_KEEP_WEBKIT_PROCESS = "markdown-keep-webkit-process"
    MARKDOWN_TEX_SUPPORT = "markdown-tex-support"
    MARKDOWN_USE_MONOSPACE_FONT = "markdown-use-monospace-font"
    MARKDOWN_RENDER_MONOSPACE_FONT_RATIO = "markdown-render-proportional-to-monospace-font-ratio"
    MARKDOWN_DEFAULT_TO_RENDER = "markdown-default-to-render"

    LAST_LAUNCHED_VERSION = "last-launched-version"
    LAST_EXPORT_DIRECTORY = "last-export-directory"
    EXTRA_EXPORT_FORMATS = "extra-export-formats"
    SHOW_EXTENDED_PREFERENCES = "show-extended-preferences"

    # Retaining to for old migrations ###################################################
    MARKDOWN_SYNTAX_HIGHLIGHTING = "markdown-syntax-highlighting-enabled"
    HIDE_EDITOR_HEADERBAR = "auto-hide-editor-headerbar"
    HIDE_HEADERBAR_WHEN_FULLSCREEN = "hide-editor-headerbar-when-fullscreen"
    #####################################################################################

    _instance = None
    _app_id = None
    settings: Gio.Settings

    def __init__(self, settings: Gio.Settings) -> None:
        super().__init__()
        self.settings = settings

    def connect_changed(self, property_name: str, callback: Callable[[], None]) -> None:
        """Connect callback to changed signal on property.

        The callback is called without the setting value.

        :param str property_name: Property to connect
        :param Callable[[], None] callback: Callback
        """
        self.settings.connect(f"changed::{property_name}", lambda _k, _v: callback())

    def create_action(self, key: str) -> Gio.Action:
        """Create action on Gio.Settings object for setting.

        :param str key: Key identifier
        :return: The action
        :rtype: Gio.Action
        """
        return self.settings.create_action(key)

    def bind(
        self, key: str, object: GObject.Object, property: str, flags: Gio.SettingsBindFlags
    ) -> None:
        """Bind setting to property.

        :param str key: Setting key
        :param GObject.Object object: Destination object to bind to
        :param str property: Destination property name
        :param Gio.SettingsBindFlags: Binding flags
        """
        self.settings.bind(key, object, property, flags)

    def get_window_size(self) -> GLib.Variant:
        """Get window size.

        :return: The size
        :rtype: GLib.Variant
        """
        return self.settings.get_value(self.WINDOW_SIZE)

    def set_window_size(self, width: int, height: int) -> None:
        """Set window size.

        :param int width: Width
        :param int height: Height
        """
        g_variant = GLib.Variant("ai", (width, height))
        self.settings.set_value(self.WINDOW_SIZE, g_variant)

    @GObject.Property(type=bool, default=True)
    def first_start(self) -> bool:
        """Get whether doing first startup.

        :return: First startup
        :rtype: bool
        """
        return self.settings.get_value(self.FIRST_START)

    @first_start.setter
    def set_first_start(self, value: bool) -> None:
        """Set whether first start.

        :param bool value: New value
        """
        self.settings.set_boolean(self.FIRST_START, value)

    @GObject.Property(type=bool, default=True)
    def use_monospace_font(self) -> bool:
        """Get whether to use a monospace font.

        :return: Using monospace font
        :rtype: bool
        """
        return self.settings.get_value(self.USE_MONOSPACE_FONT)

    @use_monospace_font.setter
    def set_use_monospace_font(self, value: bool) -> None:
        """Set whether to use a monospace font.

        :param bool value: New value
        """
        self.settings.set_boolean(self.USE_MONOSPACE_FONT, value)

    @GObject.Property(type=int)
    def font_size(self) -> int:
        """Get font size.

        :return: Size
        :rtype: int
        """
        return self.settings.get_int(self.FONT_SIZE)

    @font_size.setter
    def set_font_size(self, value: int) -> None:
        """Set font size.

        :param int value: New value
        """
        self.settings.set_int(self.FONT_SIZE, value)

    @GObject.Property(type=int)
    def default_font_size(self) -> int:
        """Get default font size.

        :return: Default size
        :rtype: int
        """
        return self.settings.get_default_value(self.FONT_SIZE).get_int32()

    @GObject.Property(type=int)
    def line_length(self) -> int:
        """Get line length.

        :return: Size in pixels
        :rtype: int
        """
        return self.settings.get_int(self.LINE_LENGTH)

    @line_length.setter
    def set_line_length(self, value: int) -> None:
        """Set line length.

        :param int value: New value
        """
        self.settings.set_int(self.LINE_LENGTH, value)

    @GObject.Property(type=int)
    def default_line_length(self) -> int:
        """Get default line length.

        :return: Default length
        :rtype: int
        """
        return self.settings.get_default_value(self.LINE_LENGTH).get_int32()

    @GObject.Property(type=int)
    def line_length_max(self) -> int:
        """Get line length maximum.

        :return: Size in pixels
        :rtype: int
        """
        return (
            self.settings.get_property("settings-schema")
            .get_key(self.LINE_LENGTH)
            .get_range()[1][-1]
        )

    @GObject.Property(type=bool, default=True)
    def spelling_enabled(self) -> bool:
        """Get whether spelling enabled.

        :return: Spelling enabled
        :rtype: bool
        """
        return self.settings.get_value(self.SPELLING_ENABLED)

    @spelling_enabled.setter
    def set_spelling_enabled(self, value: bool) -> None:
        """Set spelling enabled.

        :param bool value: New value
        """
        self.settings.set_boolean(self.SPELLING_ENABLED, value)

    @GObject.Property(type=str)
    def spelling_language(self) -> Optional[str]:
        """Get spelling language.

        :return: Language tag or None if empty
        :rtype: str
        """
        lang = self.settings.get_string(self.SPELLING_LANGUAGE)
        if lang.strip() == "":
            lang = None
        return lang

    @spelling_language.setter
    def set_spelling_language(self, value: str) -> None:
        """Set spelling language.

        :param str value: New value
        """
        self.settings.set_string(self.SPELLING_LANGUAGE, value)

    @GObject.Property(type=str)
    def style(self) -> str:
        """Get style.

        :return: Style
        :rtype: str
        """
        return self.settings.get_string(self.STYLE)

    @style.setter
    def set_style(self, value: str) -> None:
        """Set style.

        :param str value: New value
        """
        self.settings.set_string(self.STYLE, value)

    @GObject.Property(type=int)
    def sync_interval(self) -> int:
        """Get sync interval.

        :return: Interval
        :rtype: int
        """
        return self.settings.get_int(self.SYNC_INTERVAL)

    @sync_interval.setter
    def set_sync_interval(self, value: int) -> None:
        """Set sync interval.

        :param int value: New value
        """
        self.settings.set_int(self.SYNC_INTERVAL, value)

    @GObject.Property(type=str)
    def index_category_style(self) -> str:
        """Get the index category label style.

        :return: The style
        :rtype: str
        """
        return self.settings.get_string(self.INDEX_CATEGORY_STYLE)

    @index_category_style.setter
    def set_index_category_style(self, value: str) -> None:
        """Set the index category style.

        :param str value: New value
        """
        self.settings.set_string(self.INDEX_CATEGORY_STYLE, value)

    @GObject.Property(type=bool, default=True)
    def pin_sidebar(self) -> bool:
        """Get whether to pin the index sidebar (in large windows).

        :return: Whether to hide
        :rtype: bool
        """
        return self.settings.get_value(self.PERSIST_SIDEBAR)

    @pin_sidebar.setter
    def set_pin_sidebar(self, value: bool) -> None:
        """Set whether to pin the index sidebar (in large windows).

        :param bool value: New value
        """
        self.settings.set_boolean(self.PERSIST_SIDEBAR, value)

    @GObject.Property(type=str)
    def editor_theme(self) -> str:
        """Get editor theme.

        :return: Theme
        :rtype: str
        """
        return self.settings.get_string(self.EDITOR_THEME)

    @editor_theme.setter
    def set_editor_theme(self, value: str) -> None:
        """Set editor theme.

        :param str value: New value
        """
        self.settings.set_string(self.EDITOR_THEME, value)

    @property
    def editor_formatting_bar_visibility(self) -> Optional[HeaderBarVisibility]:
        """Get the formatting bar visibility.

        :return: Visibility, or None
        :rtype: HeaderBarVisibility, optional
        """
        setting_value = self.settings.get_string(self.EDITOR_FORMATTING_BAR_VISIBILTY)
        try:
            obj = HeaderBarVisibility(setting_value)
        except ValueError as e:
            logging.warning(f"Couldn't parse formatting bar visibility from '{setting_value}': {e}")
            return None
        else:
            return obj

    @editor_formatting_bar_visibility.setter
    def editor_formatting_bar_visibility(self, value: HeaderBarVisibility) -> None:
        """Set the formatting bar visibility.

        :param HeaderBarVisibility value: New value
        """
        self.settings.set_string(self.EDITOR_FORMATTING_BAR_VISIBILTY, str(value))

    @GObject.Property(type=bool, default=True)
    def editor_formatting_bar_visible_for_window_state(self) -> bool:
        """Get whether the formatting bar is configured visible for the window maximised state."

        :return: Whether visible
        :rtype: bool
        """
        setting_value = self.settings.get_string(self.EDITOR_FORMATTING_BAR_VISIBILTY)
        try:
            obj = HeaderBarVisibility(setting_value)
        except ValueError as e:
            logging.warning(f"Couldn't parse formatting bar visibility from '{setting_value}': {e}")
            return False
        else:
            visible = False
            if self.markdown_detect_syntax:
                if obj == HeaderBarVisibility.ALWAYS_VISIBLE:
                    visible = True
                elif obj == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
                    app = Gio.Application.get_default()
                    if not app:
                        logging.error("Couldn't get app?")
                        return False
                    window = app.get_active_window()
                    visible = not window.is_fullscreen()
            return visible

    @property
    def editor_header_bar_visibility(self) -> Optional[HeaderBarVisibility]:
        """Get the header bar visibility.

        :return: The visibility setting
        :rtype: HeaderBarVisibility
        """
        setting_value = self.settings.get_string(self.EDITOR_HEADER_BAR_VISIBILTY)
        try:
            obj = HeaderBarVisibility(setting_value)
        except ValueError as e:
            logging.warning(f"Couldn't parse header bar visibility from '{setting_value}': {e}")
            return None
        else:
            return obj

    @editor_header_bar_visibility.setter
    def editor_header_bar_visibility(self, value: HeaderBarVisibility) -> None:
        """Set the header bar visibility.

        :param HeaderBarVisibility value: New value
        """
        self.settings.set_string(self.EDITOR_HEADER_BAR_VISIBILTY, str(value))

    @GObject.Property(type=bool, default=True)
    def editor_header_bar_visible_for_window_state(self) -> bool:
        """Get whether the header bar is configured visible for the window maximised state."

        :return: Whether visible
        :rtype: bool
        """
        setting_value = self.settings.get_string(self.EDITOR_HEADER_BAR_VISIBILTY)
        try:
            obj = HeaderBarVisibility(setting_value)
        except ValueError as e:
            logging.warning(f"Couldn't parse header bar visibility from '{setting_value}': {e}")
            return False
        else:
            visible = False
            if obj == HeaderBarVisibility.ALWAYS_VISIBLE:
                visible = True
            elif obj == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
                app = Gio.Application.get_default()
                if not app:
                    logging.error("Couldn't get app?")
                    return False
                window = app.get_active_window()
                visible = not window.is_fullscreen()
            return visible

    @GObject.Property(type=bool, default=True)
    def markdown_render_enabled(self) -> bool:
        """Get markdown render is enabled.

        :return: Markdown render enabled
        :rtype: bool
        """
        return self.settings.get_value(self.MARKDOWN_RENDER)

    @markdown_render_enabled.setter
    def set_markdown_render_enabled(self, value: bool) -> None:
        """Set markdown render enabled.

        :param bool value: New value
        """
        self.settings.set_boolean(self.MARKDOWN_RENDER, value)

    @GObject.Property(type=bool, default=True)
    def markdown_detect_syntax(self) -> bool:
        """Get markdown syntax detection is enabled.

        :return: Enabled
        :rtype: bool
        """
        return self.settings.get_value(self.MARKDOWN_DETECT_SYNTAX)

    @markdown_detect_syntax.setter
    def set_markdown_detect_syntax(self, value: bool) -> None:
        """Set markdown syntax detection is enabled.

        :param bool value: New value
        """
        self.settings.set_boolean(self.MARKDOWN_DETECT_SYNTAX, value)

    @GObject.Property(type=bool, default=True)
    def markdown_keep_webkit_process(self) -> bool:
        """Get markdown WebKit process being retained.

        :return: WebKit process being retained
        :rtype: bool
        """
        return self.settings.get_value(self.MARKDOWN_KEEP_WEBKIT_PROCESS)

    @markdown_keep_webkit_process.setter
    def set_markdown_keep_webkit_process(self, value: bool) -> None:
        """Set markdown WebKit process being retained.

        :param bool value: New value
        """
        self.settings.set_boolean(self.MARKDOWN_KEEP_WEBKIT_PROCESS, value)

    @GObject.Property(type=bool, default=True)
    def markdown_tex_support(self) -> bool:
        """Get whether markdown TeX rendering is supported.

        :return: Markdown TeX supported
        :rtype: bool
        """
        return self.settings.get_value(self.MARKDOWN_TEX_SUPPORT)

    @markdown_tex_support.setter
    def set_markdown_tex_support(self, value: bool) -> None:
        """Set whether markdown TeX rendering is supported.

        :param bool value: New value
        """
        self.settings.set_boolean(self.MARKDOWN_TEX_SUPPORT, value)

    @GObject.Property(type=bool, default=True)
    def markdown_use_monospace_font(self) -> bool:
        """Get whether to use a monospace font for the markdown render.

        :return: Using monospace font
        :rtype: bool
        """
        return self.settings.get_value(self.MARKDOWN_USE_MONOSPACE_FONT)

    @markdown_use_monospace_font.setter
    def set_markdown_use_monospace_font(self, value: bool) -> None:
        """Set whether to use a monospace font for the markdown render.

        :param bool value: New value
        """
        self.settings.set_boolean(self.MARKDOWN_USE_MONOSPACE_FONT, value)

    @GObject.Property(type=float)
    def markdown_render_monospace_font_ratio(self) -> float:
        """Get the adjustment in size from proportional to fixed width font.

        :return: Ratio
        :rtype: float
        """
        return self.settings.get_double(self.MARKDOWN_RENDER_MONOSPACE_FONT_RATIO)

    @markdown_render_monospace_font_ratio.setter
    def set_markdown_render_monospace_font_ratio(self, value: float) -> None:
        """Set the adjustment in size from proportional to fixed width font.

        :param bool value: New value
        """
        self.settings.set_double(self.MARKDOWN_RENDER_MONOSPACE_FONT_RATIO, value)

    @GObject.Property(type=bool, default=True)
    def markdown_default_to_render(self) -> bool:
        """Get whether to render the markdown when opening the note.

        :return: Defaulting to render
        :rtype: bool
        """
        return self.settings.get_value(self.MARKDOWN_DEFAULT_TO_RENDER)

    @markdown_default_to_render.setter
    def set_markdown_default_to_render(self, value: bool) -> None:
        """Set whether to render the markdown when opening the note.

        :param bool value: New value
        """
        self.settings.set_boolean(self.MARKDOWN_DEFAULT_TO_RENDER, value)

    @GObject.Property(type=str)
    def nextcloud_endpoint(self) -> str:
        """Get Nextcloud endpoint.

        :return: Endpoint
        :rtype: str
        """
        return self.settings.get_string(self.NEXTCLOUD_ENDPOINT)

    @nextcloud_endpoint.setter
    def set_nextcloud_endpoint(self, value: str) -> None:
        """Set Nextcloud endpoint.

        :param str value: New value
        """
        self.settings.set_string(self.NEXTCLOUD_ENDPOINT, value)

    @GObject.Property(type=str)
    def nextcloud_username(self) -> str:
        """Get Nextcloud username.

        :return: Username
        :rtype: str
        """
        return self.settings.get_string(self.NEXTCLOUD_USERNAME)

    @nextcloud_username.setter
    def set_nextcloud_username(self, value: str) -> None:
        """Set Nextcloud username.

        :param str value: New value
        """
        self.settings.set_string(self.NEXTCLOUD_USERNAME, value)

    @GObject.Property(type=bool, default=False)
    def nextcloud_sync_configured(self) -> bool:
        """Get whether sync with Nextcloud is configured.

        This does not mean authentication has been successful this session.

        :return: Whether configured
        :rtype: bool
        """
        sync_username = self.nextcloud_username
        sync_endpoint = self.nextcloud_endpoint
        return sync_username != "" and sync_endpoint != ""

    @GObject.Property(type=int)
    def nextcloud_prune_threshold(self) -> int:
        """Get Nextcloud prune threshold.

        :return: Threshold
        :rtype: int
        """
        return self.settings.get_int(self.NEXTCLOUD_PRUNE_THRESHOLD)

    @nextcloud_prune_threshold.setter
    def set_nextcloud_prune_threshold(self, value: int) -> None:
        """Set Nextcloud prune threshold.

        :param int value: New value
        """
        self.settings.set_int(self.NEXTCLOUD_PRUNE_THRESHOLD, value)

    @GObject.Property(type=bool, default=True)
    def show_startup_secret_service_failure(self) -> bool:
        """Get to show Secret Service failure at startup.

        :return: Whether to show
        :rtype: bool
        """
        return self.settings.get_value(self.SHOW_STARTUP_SECRET_SERVICE_FAILURE)

    @GObject.Property(type=bool, default=True)
    def show_syncing_debug_notification(self) -> bool:
        """Get whether to show sync debug notifications.

        :return: Whether to show
        :rtype: bool
        """
        return self.settings.get_value(self.SHOW_SYNCING_DEBUG_NOTIFICATION)

    @GObject.Property(type=str)
    def backup_note_extension(self) -> str:
        """Get file extension for backed up notes.

        :return: Extension
        :rtype: str
        """
        return self.settings.get_string(self.BACKUP_NOTE_EXTENSION)

    @backup_note_extension.setter
    def set_backup_note_extension(self, value: str) -> None:
        """Set file extension for backed up notes.

        :param str value: New value
        """
        self.settings.set_string(self.BACKUP_NOTE_EXTENSION, value)

    @GObject.Property(type=str)
    def last_launched_version(self) -> str:
        """Get the last version which was run.

        :return: Version
        :rtype: str
        """
        return self.settings.get_string(self.LAST_LAUNCHED_VERSION)

    @last_launched_version.setter
    def set_last_launched_version(self, value: str) -> None:
        """Set the last version which was run.

        :param str value: New value
        """
        self.settings.set_string(self.LAST_LAUNCHED_VERSION, value)

    @GObject.Property(type=str)
    def last_export_directory(self) -> str:
        """Get the last export directory.

        :return: Directory
        :rtype: str
        """
        return self.settings.get_string(self.LAST_EXPORT_DIRECTORY)

    @last_export_directory.setter
    def set_last_export_directory(self, value: str) -> None:
        """Set the last export directory.

        :param str value: New value
        """
        self.settings.set_string(self.LAST_EXPORT_DIRECTORY, value)

    @GObject.Property(type=str)
    def extra_export_formats(self) -> str:
        """Get any extra export formats.

        :return: Raw formats string
        :rtype: str
        """
        return self.settings.get_string(self.EXTRA_EXPORT_FORMATS)

    @extra_export_formats.setter
    def set_extra_export_formats(self, value: str) -> None:
        """Set extra export formats.

        :param str value: New value
        """
        self.settings.set_string(self.EXTRA_EXPORT_FORMATS, value)

    @GObject.Property(type=bool, default=True)
    def show_extended_preferences(self) -> bool:
        """Get whether to show a great many preferences.

        :return: Whether to show
        :rtype: bool
        """
        return self.settings.get_value(self.SHOW_EXTENDED_PREFERENCES)

    @show_extended_preferences.setter
    def set_show_extended_preferences(self, value: bool) -> None:
        """Set whether to show a great many preferences.

        :param bool value: New value
        """
        self.settings.set_boolean(self.SHOW_EXTENDED_PREFERENCES, value)

    @GObject.Property(type=float)
    def network_timeout(self) -> float:
        """Get the network timeout.

        :return: Timeout in seconds
        :rtype: float
        """
        return self.settings.get_double(self.NETWORK_TIMEOUT)

    @network_timeout.setter
    def set_network_timeout(self, value: float) -> None:
        """Set the network timeout.

        :param int value: New value
        """
        self.settings.set_double(self.NETWORK_TIMEOUT, value)

    # Piped through the configuration manager for now as a result of the HTML generator using it
    # directly as configuration.
    @GObject.Property(type=bool, default=True)
    def high_contrast(self) -> bool:
        """Get whether high contrast output should be used for screen rendering.

        :return: Whether visible
        :rtype: bool
        """
        return Adw.StyleManager.get_default().get_high_contrast()

    # TODO remove ~early 2025 #####################################################

    def get_markdown_syntax_hightlighting_enabled(self) -> bool:
        """Get markdown syntax highlighting is enabled.

        :return: Highlighting enabled
        :rtype: bool
        """
        return self.settings.get_value(self.MARKDOWN_SYNTAX_HIGHLIGHTING)

    def get_hide_editor_headerbar(self) -> bool:
        """Get whether to hide the editor headerbar.

        :return: Whether to hide
        :rtype: bool
        """
        return self.settings.get_value(self.HIDE_EDITOR_HEADERBAR)

    def get_hide_editor_headerbar_when_fullscreen(self) -> bool:
        """Get whether to hide the editor headerbar when fullscreen.

        :return: Whether to hide
        :rtype: bool
        """
        return self.settings.get_value(self.HIDE_HEADERBAR_WHEN_FULLSCREEN)

    ###############################################################################

    @staticmethod
    def set_app_id(app_id: str) -> None:
        if not ConfigManager._instance:
            ConfigManager._app_id = app_id

    @staticmethod
    def get_default() -> ConfigManager:
        if not ConfigManager._instance:
            settings = Gio.Settings.new(ConfigManager._app_id)
            ConfigManager._instance = ConfigManager(settings)
        return ConfigManager._instance



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gi.repository import Adw, Gio, GLib, GObject, Gtk

from iotas.formatter import LinkStateInfo


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/link_dialog.ui")
class LinkDialog(Adw.Dialog):
    """Link editing/creating dialog."""

    __gtype_name__ = "LinkDialog"

    __gsignals__ = {
        "apply": (GObject.SignalFlags.RUN_FIRST, None, (LinkStateInfo,)),
    }

    _url_entry: Adw.EntryRow = Gtk.Template.Child()
    _text_entry: Adw.EntryRow = Gtk.Template.Child()
    _button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("submit")
        action.connect("activate", lambda _a, _p: self.__apply())
        action_group.add_action(action)
        app.set_accels_for_action("link-dialog.submit", ["<Alt>Return"])

        app.get_active_window().insert_action_group("link-dialog", action_group)
        self.__action_group = action_group

        self.connect("closed", lambda _o: self.__cleanup())

    def show(self, info: LinkStateInfo, window: Gtk.Window) -> None:
        """Show dialog.

        :param LinkStateInfo info: Link state in buffer
        :param Gtk.Window window: Parent window
        """
        if info.creating:
            # Translators: Title
            title = _("Insert Link")
            # Translators: Button
            button_label = _("Create")
        else:
            # Translators: Title
            title = _("Edit Link")
            # Translators: Button
            button_label = _("Apply")
        self._url_entry.set_text(info.link)
        self._text_entry.set_text(info.text)
        self.set_title(title)
        self._button.set_label(button_label)
        self.__info = info

        for action in self.__action_group.list_actions():
            self.__action_group.lookup_action(action).set_enabled(True)

        self.present(window)

        # Can't directly call grab_focus from idle_add otherwise its returning of True will retain,
        # and repeat, the event
        def focus_url_entry():
            self._url_entry.grab_focus()

        GLib.idle_add(focus_url_entry)

    @Gtk.Template.Callback()
    def _on_url_activated(self, _entry: Gtk.Entry) -> None:
        self._text_entry.grab_focus()

    @Gtk.Template.Callback()
    def _on_text_activated(self, _entry: Gtk.Entry) -> None:
        self.__apply()

    @Gtk.Template.Callback()
    def _on_button_clicked(self, _button: Gtk.Button) -> None:
        self.__apply()

    def __apply(self) -> None:
        self.__info.link = self._url_entry.get_text()
        self.__info.text = self._text_entry.get_text()
        self.emit("apply", self.__info)
        self.close()

    def __cleanup(self) -> None:
        for action in self.__action_group.list_actions():
            self.__action_group.lookup_action(action).set_enabled(False)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("Gtk", "4.0")
from gi.repository import Gio, GObject, Gtk

from typing import Any

from iotas.category import Category, CategorySpecialPurpose


class CategoryListModel(Gtk.SortListModel):
    ALL_INDEX = 0
    UNCATEGORISED_INDEX = 1

    def __init__(self) -> None:
        self.__all_category: Category
        self.__uncategorised_category: Category
        self.__store = Gio.ListStore(item_type=Category)
        self.__sorter = Gtk.CustomSorter()
        super().__init__(model=self.__store, sorter=self.__sorter)

    def populate(self, categories: list[Category]) -> None:
        """Populate store with categories.

        :param list[Category] categories: Categories to populate
        """
        self.__store.remove_all()
        # Populate reversed so children are created before parents
        for category in reversed(categories):
            self.__store.append(category)
            if category.special_purpose == CategorySpecialPurpose.ALL:
                self.all_category = category
            elif category.special_purpose == CategorySpecialPurpose.UNCATEGORISED:
                self.uncategorised_category = category
        self.invalidate_sort()

    def add(self, category: Category) -> None:
        """Add category to store.

        :param Category category: Category to add
        """
        index_before = 1
        for i in range(2, self.__store.get_n_items()):
            list_cat = self.__store.get_item(i)
            if list_cat.special_purpose is not None:
                continue
            if list_cat.name.lower() > category.name.lower():
                break
            index_before = i
        self.__store.insert(index_before + 1, category)

    def remove(self, category: Category) -> None:
        """Remove category from store.

        :param Category category: Category to remove
        """
        for i in range(0, self.__store.get_n_items()):
            list_cat = self.__store.get_item(i)
            if list_cat.special_purpose is not None:
                continue
            if list_cat.name == category.name:
                self.__store.remove(i)
                break

    def get_special(self, purpose: CategorySpecialPurpose) -> tuple[Category, int]:
        """Fetch a special purpose category.

        :param CategorySpecialPurpose purpose: The category a note has been removed from
        :return: The category and index in store
        :rtype: tuple[Category, int]:
        """
        category = None
        index = -1
        if purpose == CategorySpecialPurpose.ALL:
            category = self.all_category
            index = self.ALL_INDEX
        elif purpose == CategorySpecialPurpose.UNCATEGORISED:
            category = self.uncategorised_category
            index = self.UNCATEGORISED_INDEX
        return (category, index)

    def get_non_special(self, name: str) -> tuple[Category | None, int]:
        """Fetch a non-special purpose category.

        :param str name: The category name
        :return: The category and index in store
        :rtype: tuple[Category, int]:
        """
        category = None
        index = 0
        for i in range(0, self.__store.get_n_items()):
            list_cat = self.__store.get_item(i)
            if list_cat.name == name:
                category = list_cat
                index = i
                break
        return (category, index)

    def invalidate_sort(self) -> None:
        """Invalidate the sorter."""
        self.__sorter.set_sort_func(CategoryListModel.sort_func)

    @GObject.Property(type=Category)
    def all_category(self) -> Category:
        return self.__all_category

    @all_category.setter
    def set_all_category(self, new_val: Category) -> None:
        self.__all_category = new_val

    @GObject.Property(type=Category)
    def uncategorised_category(self) -> Category:
        return self.__uncategorised_category

    @uncategorised_category.setter
    def set_uncategorised_category(self, new_val: Category) -> None:
        self.__uncategorised_category = new_val

    @staticmethod
    def sort_func(category1: Category, category2: Category, _data: Any) -> int:
        """Sort categories.

        :param Category category1: First category
        :param Category category2: Second category
        :param _data: Unused
        :return: -1 if category1 earlier, 1 if later, 0 unused
        :rtype: int
        """
        if category1.special_purpose == CategorySpecialPurpose.ALL:
            return -1
        elif category2.special_purpose == CategorySpecialPurpose.ALL:
            return 1
        elif (
            category1.special_purpose == CategorySpecialPurpose.UNCATEGORISED
            and category2.special_purpose is None
        ):
            return -1
        elif (
            category2.special_purpose == CategorySpecialPurpose.UNCATEGORISED
            and category1.special_purpose is None
        ):
            return 1
        else:
            return -1 if category1.name.lower() < category2.name.lower() else 1



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import shutil
import unittest
import warnings

from iotas.attachment import Attachment
from iotas.attachment_helpers import (
    copy_note_attachments,
    write_attachment_to_disk,
    get_attachments_on_disk_for_note,
    delete_attachments_for_note,
    get_attachment_disk_states,
    get_attachment_filesystem_uri,
    get_attachment_filesystem_path,
    attachment_exists_locally,
    get_attachment_file_object,
    flush_all_attachments_on_disk,
    set_attachment_filesystem_uris_on_tokens,
    trim_orphaned_images,
    AttachmentsCopyOutcome,
)
from iotas.markdown_helpers import (
    filter_image_tokens,
    parse_to_tokens,
)
from iotas.note import Note

import iotas.attachment_helpers


def get_attachments_dir() -> str:
    file_dir = os.path.dirname(__file__)
    return os.path.join(file_dir, os.pardir, "testing-tmp", "user-attachments")


iotas.attachment_helpers.get_attachments_dir = get_attachments_dir

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_copy_note_attachments(self) -> None:
        self.__prepare_output_dir()
        note = Note()
        note.id = 1
        note.content = "![](image.png)\n\n![](image2.png)\n\n![](image2.png)"
        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "1.image2.png")
        self.assertTrue(self.__create_empty_file(path))

        out_dir = os.path.join(get_attachments_dir(), "copy-test")

        def refresh_out_dir():
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)
            os.makedirs(out_dir)

        refresh_out_dir()
        results = copy_note_attachments(note, out_dir, prefix_note_id=True)
        self.assertEqual(results.outcome, AttachmentsCopyOutcome.SUCCESS)
        self.assertEqual(len(os.listdir(out_dir)), 2)

        refresh_out_dir()
        note.content = "![](image.png)\n\n![](image2.png)\n\n![](image3.png)"
        results = copy_note_attachments(note, out_dir, prefix_note_id=True)
        self.assertEqual(results.outcome, AttachmentsCopyOutcome.HAD_MISSING)
        self.assertEqual(len(os.listdir(out_dir)), 2)

        refresh_out_dir()
        note.content = "[](image.png)"
        results = copy_note_attachments(note, out_dir, prefix_note_id=True)
        self.assertEqual(results.outcome, AttachmentsCopyOutcome.NONE)
        self.assertEqual(len(os.listdir(out_dir)), 0)

        self.__clean_output_dir()

    def test_write_attachment_to_disk(self) -> None:
        self.__prepare_output_dir()
        attachment = Attachment()
        attachment.note_id = 1
        attachment.path = "image.png"

        success = write_attachment_to_disk(attachment, b"bytes")
        self.assertTrue(success)
        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(os.path.exists(path))
        with open(path, "rb") as f:
            out = f.read()
        self.assertEqual(out, b"bytes")

        self.__clean_output_dir()

    def test_get_attachments_on_disk_for_note(self) -> None:
        self.__prepare_output_dir()
        note = Note()
        note.id = 1
        self.assertEqual(len(get_attachments_on_disk_for_note(note)), 0)

        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "1.image2.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "2.image.png")
        self.assertTrue(self.__create_empty_file(path))

        out = get_attachments_on_disk_for_note(note)
        self.assertEqual(len(out), 2)
        out_paths = {a.path for a in out}
        self.assertEqual(out_paths, {"image.png", "image2.png"})

        self.__clean_output_dir()

    def test_delete_attachments_for_note(self) -> None:
        self.__prepare_output_dir()
        note = Note()
        note.id = 1
        self.assertEqual(len(get_attachments_on_disk_for_note(note)), 0)

        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "1.image2.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "2.image.png")
        self.assertTrue(self.__create_empty_file(path))

        delete_attachments_for_note(note)
        remaining = os.listdir(get_attachments_dir())
        self.assertEqual(len(remaining), 1)
        self.assertEqual(remaining[0], "2.image.png")

        self.__clean_output_dir()

    def test_get_attachment_disk_states(self) -> None:
        self.__prepare_output_dir()

        note = Note()
        note.id = 1
        note.content = "![](image.png)\n\n![](image2.png)\n\n![](image2.png)\n\n![](image3.png)"

        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "1.image2.png")
        self.assertTrue(self.__create_empty_file(path))

        _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        states = get_attachment_disk_states(note, tokens)
        self.assertEqual(len(states.exists), 2)
        self.assertTrue("image.png" in states.exists)
        self.assertTrue("image2.png" in states.exists)
        self.assertEqual(len(states.missing), 1)
        self.assertTrue("image3.png" in states.missing)
        self.__clean_output_dir()

    def test_get_attachment_filesystem_uri(self) -> None:
        attachment = Attachment()
        attachment.note_id = 1
        attachment.path = "image.png"
        uri = get_attachment_filesystem_uri(attachment)
        expected = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertEqual(uri, f"file://{expected}")

    def test_get_attachment_filesystem_path(self) -> None:
        attachment = Attachment()
        attachment.note_id = 1
        attachment.path = "image.png"
        uri = get_attachment_filesystem_path(attachment)
        expected = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertEqual(uri, expected)

    def test_attachment_exists_locally(self) -> None:
        attachment = Attachment()
        attachment.note_id = 1
        attachment.path = "image.png"
        self.assertFalse(attachment_exists_locally(attachment))

        self.__prepare_output_dir()
        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))
        self.assertTrue(attachment_exists_locally(attachment))
        self.__clean_output_dir()

    def test_get_attachment_file_object(self) -> None:
        attachment = Attachment()
        attachment.note_id = 1
        attachment.path = "image.png"
        self.__prepare_output_dir()
        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))

        obj = get_attachment_file_object(attachment)
        self.assertIsNotNone(obj)
        if obj is not None:  # mypy
            self.assertEqual(obj.read(), b"")
        self.__clean_output_dir()

    def test_flush_all_attachments_on_disk(self) -> None:
        self.__prepare_output_dir()
        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "1.image2.png")
        self.assertTrue(self.__create_empty_file(path))
        flush_all_attachments_on_disk()
        self.assertEqual(len(os.listdir(get_attachments_dir())), 0)
        self.__clean_output_dir()

    def test_set_attachment_filesystem_uris_to_tokens(self) -> None:
        note = Note()
        note.id = 1
        note.content = "![](image.png)\n\n![](image2.png)\n\n![](image2.png)"
        _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)

        attachment1 = Attachment()
        attachment1.note_id = 1
        attachment1.path = "image.png"
        attachment2 = Attachment()
        attachment2.note_id = 1
        attachment2.path = "image2.png"
        new_paths = {
            "image.png": attachment1,
            "image2.png": attachment2,
        }
        set_attachment_filesystem_uris_on_tokens(tokens, new_paths)
        image_tokens = filter_image_tokens(tokens)
        self.assertEqual(len(image_tokens), 3)

        dirpath = f"file://{os.path.join(get_attachments_dir())}/"
        path = dirpath + "1.image.png"
        self.assertEqual(image_tokens[0].attrs["src"], path)
        path = dirpath + "1.image2.png"
        self.assertEqual(image_tokens[1].attrs["src"], path)
        self.assertEqual(image_tokens[2].attrs["src"], path)

    def test_trim_orphaned_images(self) -> None:
        self.__prepare_output_dir()
        path = os.path.join(get_attachments_dir(), "1.image.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "1.orphan1.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "1.orphan2.png")
        self.assertTrue(self.__create_empty_file(path))
        path = os.path.join(get_attachments_dir(), "2.image.png")
        self.assertTrue(self.__create_empty_file(path))
        self.assertEqual(len(os.listdir(get_attachments_dir())), 4)

        note = Note()
        note.id = 1
        note.content = "![description](image.png)"
        trim_orphaned_images(note, on_thread=False)
        self.assertEqual(len(os.listdir(get_attachments_dir())), 2)

    def __prepare_output_dir(self) -> str:
        out_path = get_attachments_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)
        os.makedirs(out_path)
        return out_path

    def __clean_output_dir(self) -> None:
        out_path = get_attachments_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)

    def __create_empty_file(self, path: str) -> bool:
        try:
            with open(path, "w") as f:
                f.write("")
        except OSError:
            return False
        else:
            return True



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("GtkSource", "5")
from gi.repository import Adw, GObject, GtkSource, Pango

import logging


class FocusModeHelper(GObject.Object):

    # Adwaita light_4
    UNFOCUSED_COLOUR_LIGHT = "#c0bfbc"
    # Adwaita dark_2
    UNFOCUSED_COLOUR_DARK = "#5e5c64"

    def __init__(self, buffer: GtkSource.Buffer) -> None:
        super().__init__()
        self.__buffer = buffer
        self.__handler_ids: list[int] = []
        self.__active = False

        start_iter = buffer.get_start_iter()
        self.__last_start = buffer.create_mark(None, start_iter)
        self.__last_end = buffer.create_mark(None, start_iter)

        style_manager = Adw.StyleManager.get_default()
        style_manager.connect("notify::dark", self.__on_style_change)
        init_colour = (
            self.UNFOCUSED_COLOUR_DARK if style_manager.get_dark() else self.UNFOCUSED_COLOUR_LIGHT
        )

        self.__tag = buffer.create_tag(
            "unfocused", foreground=init_colour, underline=Pango.Underline.NONE
        )

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value
        if self.__active and len(self.__handler_ids) == 0:
            self.__activate()
        elif not self.__active and len(self.__handler_ids) > 0:
            self.__deactivate()

    def __activate(self) -> None:
        logging.debug("Activating focus mode")
        for signal in ("cursor-moved", "changed"):
            handler_id = self.__buffer.connect(signal, lambda _o: self.__refresh(initial=False))
            self.__handler_ids.append(handler_id)
        self.__reset_marks()
        self.__refresh(initial=True)

    def __deactivate(self) -> None:
        logging.debug("Deactivating focus mode")
        self.__buffer.remove_tag(
            self.__tag, self.__buffer.get_start_iter(), self.__buffer.get_end_iter()
        )
        for handler_id in self.__handler_ids:
            self.__buffer.disconnect(handler_id)
        self.__handler_ids = []

    def __refresh(self, initial: bool) -> None:
        buffer = self.__buffer

        # Determine current sentence
        start_iter = buffer.get_iter_at_mark(buffer.get_insert())
        if not start_iter.starts_sentence() and not start_iter.starts_line():
            start_iter.backward_sentence_start()
        end_iter = start_iter.copy()
        if not start_iter.ends_line():
            end_iter.forward_sentence_end()

        # Check if changed
        last_start = buffer.get_iter_at_mark(self.__last_start)
        last_end = buffer.get_iter_at_mark(self.__last_end)
        start_changed = not last_start.equal(start_iter) or initial
        end_changed = not last_end.equal(end_iter) or initial
        if not start_changed and not end_changed:
            return

        # Untag last sections
        if start_changed:
            buffer.remove_tag(self.__tag, buffer.get_start_iter(), last_start)
        if end_changed:
            buffer.remove_tag(self.__tag, last_end, buffer.get_end_iter())

        # Tag new sections
        if start_changed:
            if not start_iter.is_start():
                buffer.apply_tag(self.__tag, buffer.get_start_iter(), start_iter)
            buffer.move_mark(self.__last_start, start_iter)
        if end_changed:
            if not end_iter.is_end():
                buffer.apply_tag(self.__tag, end_iter, buffer.get_end_iter())
            buffer.move_mark(self.__last_end, end_iter)

    def __on_style_change(
        self,
        style_manager: Adw.StyleManager,
        _value: GObject.ParamSpec,
    ) -> None:
        if style_manager.get_dark():
            self.__tag.set_property("foreground", self.UNFOCUSED_COLOUR_DARK)
        else:
            self.__tag.set_property("foreground", self.UNFOCUSED_COLOUR_LIGHT)

    def __reset_marks(self) -> None:
        start_iter = self.__buffer.get_start_iter()
        self.__buffer.move_mark(self.__last_start, start_iter)
        self.__buffer.move_mark(self.__last_end, start_iter)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import unittest
import warnings

from iotas.html_generator_configuration import HtmlGeneratorConfiguration

from iotas.markdown_helpers import parse_to_tokens
from iotas.note import Note
from iotas.template_html_generator import TemplateHtmlGenerator, TemplateDetails

warnings.filterwarnings("ignore", "version")


class _TestHtmlGeneratorConfig(HtmlGeneratorConfiguration):

    tex_support = False

    @property
    def markdown_tex_support(self) -> bool:
        """Get whether markdown TeX rendering is supported.

        :return: Markdown TeX supported
        :rtype: bool
        """
        return self.tex_support

    @property
    def line_length(self) -> int:
        """Get line length.

        :return: Size in pixels
        :rtype: int
        """
        return 800

    @property
    def markdown_use_monospace_font(self) -> bool:
        """Get whether to use a monospace font for the markdown render.

        :return: Using monospace font
        :rtype: bool
        """
        return False

    @property
    def use_monospace_font(self) -> bool:
        """Get whether to use a monospace font.

        :return: Using monospace font
        :rtype: bool
        """
        return True

    @property
    def font_size(self) -> int:
        """Get font size.

        :return: Size
        :rtype: int
        """
        return 14

    @property
    def high_contrast(self) -> bool:
        """Get whether high contrast output should be used for screen rendering.

        :return: Whether visible
        :rtype: bool
        """
        return False

    @property
    def markdown_render_monospace_font_ratio(self) -> float:
        """Get the adjustment in size from proportional to fixed width font.

        :return: Ratio
        :rtype: float
        """
        return 1.0

    @property
    def editor_header_bar_visible_for_window_state(self) -> bool:
        """Get whether the header bar is configured visible for the window maximised state."

        :return: Whether visible
        :rtype: bool
        """
        return False


class Test(unittest.TestCase):

    def test_generate(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.id = 1
        note.title = "Test Note Title"

        generator = self.__reset()
        parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        html = generator.generate(
            note,
            tokens,
            parser.renderer.render,
            parser.options,
            searching=False,
            export_format=None,
        )
        self.assertTrue(note.content in html)
        self.assertTrue(note.title in html)
        self.assertFalse("katex" in html)
        self.assertFalse("<style>" in html)
        self.assertFalse("scrollTop = " in html)
        self.assertTrue("/mock/data/path" in html)

        generator = self.__reset(tex_support=True)
        parser, tokens = parse_to_tokens(note, exporting=False, tex_support=True)
        html = generator.generate(
            note,
            tokens,
            parser.renderer.render,
            parser.options,
            searching=False,
            export_format=None,
        )
        self.assertTrue("katex" in html)

        generator = self.__reset()
        parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        html = generator.generate(
            note,
            tokens,
            parser.renderer.render,
            parser.options,
            searching=False,
            export_format=None,
            scroll_position=100.0,
        )
        self.assertTrue("scrollTop = 100.0" in html)

        note.content = "![title](https://imageurl.com/image.png)"
        generator = self.__reset()
        parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        html = generator.generate(
            note,
            tokens,
            parser.renderer.render,
            parser.options,
            searching=False,
            export_format=None,
        )
        self.assertTrue('loading="lazy"' in html)

        note.content = "![](image.png)"
        generator = self.__reset()
        parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        html = generator.generate(
            note,
            tokens,
            parser.renderer.render,
            parser.options,
            searching=False,
            export_format=None,
        )
        self.assertTrue('src="image.png"' in html)

        generator = self.__reset()
        parser, tokens = parse_to_tokens(note, exporting=True, tex_support=False)
        html = generator.generate(
            note,
            tokens,
            parser.renderer.render,
            parser.options,
            searching=False,
            export_format="html",
        )
        self.assertTrue('src="attachments/image.png"' in html)

        generator = self.__reset()
        parser, tokens = parse_to_tokens(note, exporting=True, tex_support=False)
        html = generator.generate(
            note,
            tokens,
            parser.renderer.render,
            parser.options,
            searching=False,
            export_format="html",
        )
        self.assertFalse("scrollTop = " in html)
        self.assertTrue("font-size: 14pt" in html)
        self.assertTrue("max-width: 800px" in html)
        self.assertTrue("padding-top: 47px" in html)
        self.assertFalse("::selection {" in html)
        self.assertFalse("/mock/data/path" in html)
        self.assertFalse('loading="lazy"' in html)

    def test_generate_user_stylesheet(self) -> None:
        generator = self.__reset()
        css = generator.generate_user_stylesheet(searching=False, height=500)
        self.assertTrue("font-size: 14pt" in css)
        self.assertTrue("max-width: 800px" in css)
        self.assertTrue("padding-top: 47px" in css)
        self.assertTrue("padding-bottom: 200px" in css)
        self.assertFalse("::selection {" in css)

        css = generator.generate_user_stylesheet(searching=True, height=500)
        self.assertTrue("::selection {" in css)

    def test_update_font_family(self) -> None:
        generator = self.__reset()
        generator.update_font_family("unique")
        css = generator.generate_user_stylesheet(searching=False, height=500)
        self.assertTrue("font-family: unique" in css)

    def __reset(self, tex_support=False) -> TemplateHtmlGenerator:
        config = _TestHtmlGeneratorConfig()
        config.tex_support = tex_support
        data_path = "/mock/data/path"
        header_bar_height = 47
        return TemplateHtmlGenerator(config, data_path, header_bar_height, self.__template_loader)

    def __template_loader(self) -> TemplateDetails:
        file_dir = os.path.dirname(__file__)
        templates_path = os.path.join(file_dir, os.pardir, "data", "render-template")
        with open(os.path.join(templates_path, "template.html"), "r") as f:
            html = f.read()
        with open(os.path.join(templates_path, "binding.js"), "r") as f:
            js = f.read()
        with open(os.path.join(templates_path, "configurable.css"), "r") as f:
            css = f.read()
        with open(os.path.join(templates_path, "searching.css"), "r") as f:
            search_css = f.read()
        return TemplateDetails(html, js, css, search_css)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk

from typing import Any

from iotas.outline_generator import OutlineHeading
from iotas.ui_utils import check_for_search_starting, check_for_open_first_result_shortcut


class _HeadingRow(Adw.ActionRow):
    """Row for outline dialog headings.

    :param str label: Text
    :param int line: Markdown line number
    """

    def __init__(self, label: str, line: int) -> None:
        super().__init__()
        self.set_title(label)
        self.line = line
        self.set_activatable(True)


class _SizeReportingListbox(Gtk.ListBox):

    __gsignals__ = {
        "new-height": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
    }

    def __init__(self) -> None:
        super().__init__()
        self.set_selection_mode(Gtk.SelectionMode.NONE)
        self.set_activate_on_single_click(True)
        self.add_css_class("boxed-list")

    def do_size_allocate(self, width: int, height: int, baseline: int) -> None:
        Gtk.ListBox.do_size_allocate(self, width, height, baseline)
        self.emit("new-height", height)


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/outline_dialog.ui")
class OutlineDialog(Adw.Dialog):
    """Note outline dialog."""

    __gtype_name__ = "OutlineDialog"

    __gsignals__ = {
        "jump-to": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
    }

    _header_stack: Gtk.Stack = Gtk.Template.Child()
    _main_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _filter_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _stack: Gtk.Stack = Gtk.Template.Child()
    _spinner: Adw.Spinner = Gtk.Template.Child()
    _list: Gtk.ScrolledWindow = Gtk.Template.Child()
    _no_headings: Adw.StatusPage = Gtk.Template.Child()
    _filter_entry: Gtk.SearchEntry = Gtk.Template.Child()
    _zero_filtered: Adw.StatusPage = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        self.__store = Gio.ListStore(item_type=OutlineHeading)
        self.__filter_term = ""

        self.__filter_model = Gtk.FilterListModel()
        self.__filter = Gtk.CustomFilter()
        self.__filter_model.set_filter(self.__filter)

        self.__filter.set_filter_func(self.__filter_heading)
        self.__filter_model.set_model(self.__store)

        self.__listbox = _SizeReportingListbox()
        self.__listbox.connect(
            "new-height", lambda _o, height: self.__update_for_list_height(height)
        )
        self.__listbox.connect("row-activated", lambda _o, row: self.__activate_row(row))
        self._list.set_child(self.__listbox)

        controller = Gtk.EventControllerKey()
        controller.connect("key-pressed", self.__on_listbox_key_pressed)
        self.__listbox.add_controller(controller)

        controller = Gtk.EventControllerKey()
        controller.connect("key-pressed", self.__on_entry_key_pressed)
        self._filter_entry.add_controller(controller)

        self.__listbox.bind_model(self.__filter_model, self.__create_row, None)

    def populate(self, headings: list[OutlineHeading], rtl: bool) -> None:
        """Populate headings.

        :param list[OutlineHeading] headings: Headings to populate
        :param bool rtl: Whether direction is right-to-left
        """
        self._spinner.set_visible(False)

        if not headings:
            self._stack.set_visible_child(self._no_headings)
            return

        self.__rtl = rtl

        for heading in headings:
            self.__store.append(heading)

        self._main_headerbar.add_css_class("list-matched")
        self._stack.set_visible_child(self._list)
        self.__listbox.grab_focus()

    @Gtk.Template.Callback()
    def _on_close_clicked(self, _button: Gtk.Button) -> None:
        self.close()

    @Gtk.Template.Callback()
    def _on_filter_changed(self, entry: Gtk.Editable) -> None:
        if not list(self.__store):
            return
        self.__filter_term = entry.get_text()
        self.__filter.set_filter_func(self.__filter_heading)
        if self.__filter_term != "" and not list(self.__listbox):
            self._stack.set_visible_child(self._zero_filtered)
        elif self._stack.get_visible_child() != self._list:
            self._stack.set_visible_child(self._list)

    @Gtk.Template.Callback()
    def _on_stop_filter(self, entry: Gtk.Editable) -> None:
        entry.set_text("")
        self.__filter_term = ""
        self.__filter.set_filter_func(self.__filter_heading)
        self.__listbox.grab_focus()
        self._header_stack.set_visible_child(self._main_headerbar)

    @Gtk.Template.Callback()
    def _on_entry_activate(self, entry: Gtk.Editable) -> None:
        self.__activate_first_result()

    def __on_listbox_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if self.__check_if_filter_starting(controller, keyval, state, clear=False):
            self.__show_filter()
            return Gdk.EVENT_STOP

        if not self.__listbox.is_focus():
            return Gdk.EVENT_PROPAGATE

        if keyval in (Gdk.KEY_Down, Gdk.KEY_KP_Down):
            row = self.__listbox.get_row_at_index(0)
            row.grab_focus()
            return Gdk.EVENT_STOP
        elif keyval in (Gdk.KEY_Up, Gdk.KEY_KP_Up):
            count = len(list(self.__listbox))
            row = self.__listbox.get_row_at_index(count - 1)
            row.grab_focus()
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def __on_entry_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval in (Gdk.KEY_Down, Gdk.KEY_KP_Down):
            row = self.__listbox.get_row_at_index(0)
            row.grab_focus()
            return Gdk.EVENT_STOP
        elif check_for_open_first_result_shortcut(keyval, state):
            self.__activate_first_result()

        return Gdk.EVENT_PROPAGATE

    def __update_for_list_height(self, height: int) -> None:
        window_height = self.get_root().get_height()
        threshold = window_height * 0.9
        if height > threshold:
            height = threshold
        GLib.idle_add(self._stack.set_property, "height-request", height)

    def __show_filter(self) -> None:
        self._header_stack.set_visible_child(self._filter_headerbar)
        self._filter_entry.grab_focus()

    def __check_if_filter_starting(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        state: Gdk.ModifierType,
        clear: bool = True,
    ) -> bool:
        if not check_for_search_starting(controller, keyval, state):
            return False

        if clear:
            self._filter_entry.set_text("")
        if controller.forward(self._filter_entry.get_delegate()):
            return True

        return False

    def __create_row(self, heading: OutlineHeading, _user_data: Any) -> Gtk.Widget:
        row = _HeadingRow(heading.text, heading.line)
        if heading.indent_level:
            row.add_prefix
            padding = Gtk.Label()
            padding.set_text("     " * heading.indent_level)
            if self.__rtl:
                row.add_suffix(padding)
            else:
                row.add_prefix(padding)
        return row

    def __filter_heading(self, heading: OutlineHeading) -> bool:
        return self.__filter_term.lower() in heading.text.lower()

    def __activate_row(self, row: _HeadingRow) -> None:
        self.emit("jump-to", row.line)
        self.close()

    def __activate_first_result(self) -> None:
        if list(self.__listbox):
            row = self.__listbox.get_row_at_index(0)
            self.__activate_row(row)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gi.repository import GObject, Gtk


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/editor_search_entry.ui")
class EditorSearchEntry(Gtk.Box):
    __gtype_name__ = "EditorSearchEntry"
    __gsignals__ = {
        "activate": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    text: Gtk.Text = Gtk.Template.Child()
    _info: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        self.__occurrence_count = 0
        self.__occurrence_position = 0
        self.text.connect("notify::text", lambda _o, _v: self.__on_text_changed())
        super().__init__()

    def set_occurrence_count(self, count: int) -> None:
        """Set the number of search term occurrences.

        :param int count: New value
        """
        self.__occurrence_count = count
        if count or self.text.get_text() == "":
            self.remove_css_class("error")
        else:
            self.add_css_class("error")
        self.__update_position()

    def set_occurrence_position(self, position: int) -> None:
        """Set the current search term occurrence.

        :param int position: New value
        """
        self.__occurrence_position = position
        self.__update_position()

    def select_all_and_focus(self) -> None:
        """Select and focus all text."""
        if self.text.get_text() != "":
            self.text.select_region(0, -1)
        self.text.grab_focus()

    def show_working(self) -> None:
        """Show feedback that search is running.

        Rarely necessary but can occur if searching immediately after opening large notes.
        """
        self._info.set_label("…")
        self.remove_css_class("error")

    def get_text(self) -> str:
        """Fetch the text from the entry.

        :return: Unaccented form
        :rtype: str
        """
        return self.text.get_text()

    def set_text(self, text: str) -> None:
        """Set the text on the entry.

        :param str text: New value
        """
        return self.text.set_text(text)

    @Gtk.Template.Callback()
    def _on_activate(self, _obj: GObject.Object) -> None:
        self.emit("activate")

    def __on_text_changed(self) -> None:
        if self.get_text() == "":
            self.__occurrence_count = 0
            self.__occurrence_position = 0
            self._info.set_label("")
        else:
            self._info.set_label("…")

    def __update_position(self) -> None:
        if self.__occurrence_count == 0:
            occurrence_str = ""
        else:
            # Translators: Description, {0} the current position in {1} a number of search results
            occurrence_str = _("{0} of {1}").format(
                self.__occurrence_position, self.__occurrence_count
            )
        self._info.set_label(occurrence_str)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject

from iotas.category_header_bar import CategoryHeaderBar
from iotas.editor import Editor
from iotas.editor_rename_header_bar import EditorRenameHeaderBar
from iotas.editor_search_entry import EditorSearchEntry
from iotas.editor_search_header_bar import EditorSearchHeaderBar
from iotas.editor_text_view import EditorTextView
from iotas.first_start_page import FirstStartPage
from iotas.font_size_selector import FontSizeSelector
from iotas.formatting_header_bar import FormattingHeaderBar
from iotas.index import Index
from iotas.index_menu_button import IndexMenuButton
from iotas.index_note_list import IndexNoteList
from iotas.index_search_header_bar import IndexSearchHeaderBar
from iotas.render_search_header_bar import RenderSearchHeaderBar
from iotas.selection_header_bar import SelectionHeaderBar
from iotas.sidebar import Sidebar
from iotas.sidebar_row import SidebarRow
from iotas.theme_selector import ThemeSelector


def load_widgets() -> None:
    """Ensure that the types have been registered with the type system.

    Allows them to be used in template UI files.
    """
    GObject.type_ensure(CategoryHeaderBar)
    GObject.type_ensure(Editor)
    GObject.type_ensure(EditorRenameHeaderBar)
    GObject.type_ensure(EditorSearchEntry)
    GObject.type_ensure(EditorSearchHeaderBar)
    GObject.type_ensure(EditorTextView)
    GObject.type_ensure(FirstStartPage)
    GObject.type_ensure(FontSizeSelector)
    GObject.type_ensure(FormattingHeaderBar)
    GObject.type_ensure(Index)
    GObject.type_ensure(IndexMenuButton)
    GObject.type_ensure(IndexNoteList)
    GObject.type_ensure(IndexSearchHeaderBar)
    GObject.type_ensure(RenderSearchHeaderBar)
    GObject.type_ensure(SelectionHeaderBar)
    GObject.type_ensure(Sidebar)
    GObject.type_ensure(SidebarRow)
    GObject.type_ensure(ThemeSelector)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
import warnings

from gi.repository import GLib

from iotas.outline_generator import OutlineGenerator, OutlineHeading
from iotas.note import Note

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    headings: list

    def test_generate(self) -> None:
        self.__generate("plain text")
        self.assertEqual(len(self.headings), 0)

        self.__generate("# Heading\n## Second\n### Third\nplain")
        self.assertEqual(len(self.headings), 3)
        heading = self.headings[0]
        self.assertIsInstance(heading, OutlineHeading)
        self.assertEqual(heading.text, "Heading")
        self.assertEqual(heading.line, 0)
        self.assertEqual(heading.level, 1)
        self.assertEqual(heading.indent_level, 0)
        heading = self.headings[1]
        self.assertIsInstance(heading, OutlineHeading)
        self.assertEqual(heading.text, "Second")
        self.assertEqual(heading.line, 1)
        self.assertEqual(heading.level, 2)
        self.assertEqual(heading.indent_level, 1)
        heading = self.headings[2]
        self.assertIsInstance(heading, OutlineHeading)
        self.assertEqual(heading.text, "Third")
        self.assertEqual(heading.line, 2)
        self.assertEqual(heading.level, 3)
        self.assertEqual(heading.indent_level, 2)

        # Test indentation level squashing
        self.__generate("# Heading\n### Second\n#### Third")
        self.assertEqual(len(self.headings), 3)
        heading = self.headings[0]
        self.assertIsInstance(heading, OutlineHeading)
        self.assertEqual(heading.text, "Heading")
        self.assertEqual(heading.line, 0)
        self.assertEqual(heading.level, 1)
        self.assertEqual(heading.indent_level, 0)
        heading = self.headings[1]
        self.assertIsInstance(heading, OutlineHeading)
        self.assertEqual(heading.text, "Second")
        self.assertEqual(heading.line, 1)
        self.assertEqual(heading.level, 3)
        self.assertEqual(heading.indent_level, 1)
        heading = self.headings[2]
        self.assertIsInstance(heading, OutlineHeading)
        self.assertEqual(heading.text, "Third")
        self.assertEqual(heading.line, 2)
        self.assertEqual(heading.level, 4)
        self.assertEqual(heading.indent_level, 2)

    def __generate(self, content: str) -> None:
        generator = OutlineGenerator()
        note = Note(new_note=True)
        note.content = content
        self.done = False
        context = GLib.MainContext.default()
        generator.generate(note, self.__callback)
        while not self.done:
            context.iteration(False)

    def __callback(self, headings: list[OutlineHeading]) -> None:
        self.headings = headings
        self.done = True



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import unittest
import warnings

from gi.repository import Gtk, GtkSource, GLib

from iotas import text_utils

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_sanitise_path(self) -> None:
        self.assertEqual(text_utils.sanitise_path('..*|/\\:"<>?'), "")
        self.assertEqual(text_utils.sanitise_path(""), "")

    def test_parse_any_url_at_iter(self) -> None:
        buffer = Gtk.TextBuffer()
        start_iter = buffer.get_start_iter()
        self.assertIsNone(text_utils.parse_any_url_at_iter(start_iter, http_only=True))

        buffer = Gtk.TextBuffer()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "plain")
        start_iter = buffer.get_start_iter()
        self.assertIsNone(text_utils.parse_any_url_at_iter(start_iter, http_only=True))

        buffer = Gtk.TextBuffer()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "https://test.com")
        start_iter = buffer.get_start_iter()
        self.assertIsNotNone(text_utils.parse_any_url_at_iter(start_iter, http_only=True))

        buffer = Gtk.TextBuffer()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "other://test.com")
        start_iter = buffer.get_start_iter()
        self.assertIsNone(text_utils.parse_any_url_at_iter(start_iter, http_only=True))

        buffer = Gtk.TextBuffer()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "other://test.com")
        start_iter = buffer.get_start_iter()
        self.assertIsNotNone(text_utils.parse_any_url_at_iter(start_iter, http_only=False))

    def test_has_unicode(self) -> None:
        self.assertTrue(text_utils.has_unicode("👋"))
        self.assertFalse(text_utils.has_unicode("plain"))
        self.assertFalse(text_utils.has_unicode(""))

    def test_str_is_url(self) -> None:
        self.assertTrue(text_utils.str_is_url("https://test.com", http_only=True))
        self.assertFalse(text_utils.str_is_url("httpsX://test.com", http_only=True))
        self.assertFalse(text_utils.str_is_url("https:test.com", http_only=True))
        self.assertFalse(text_utils.str_is_url("http:test.com", http_only=True))
        self.assertFalse(text_utils.str_is_url("httptest.com", http_only=True))
        self.assertTrue(text_utils.str_is_url("other://test.com", http_only=False))
        self.assertFalse(text_utils.str_is_url("plain", http_only=False))
        self.assertFalse(text_utils.str_is_url("plain", http_only=True))
        self.assertFalse(text_utils.str_is_url("", http_only=True))
        self.assertFalse(text_utils.str_is_url("", http_only=False))

    def test_iter_forward_to_context_class_removed(self) -> None:
        manager = GtkSource.LanguageManager.get_default()
        language = manager.get_language("c")

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "// comment\nint i = 0;")
        start_iter = buffer.get_start_iter()
        start_iter.forward_chars(5)
        self.assertEqual(buffer.get_context_classes_at_iter(start_iter), ["comment"])
        moved = text_utils.iter_forward_to_context_class_removed(start_iter, "comment")
        self.assertTrue(moved)
        self.assertEqual(start_iter.get_offset(), 10)

    def test_iter_backward_to_context_class_start(self) -> None:
        manager = GtkSource.LanguageManager.get_default()
        language = manager.get_language("c")

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "int i = 0;\n// comment")
        end_iter = buffer.get_end_iter()
        end_iter.backward_char()
        self.assertEqual(buffer.get_context_classes_at_iter(end_iter), ["comment"])
        moved = text_utils.iter_backward_to_context_class_start(end_iter, "comment")
        self.assertTrue(moved)
        self.assertEqual(end_iter.get_offset(), 11)

    def test_get_iters_at_sourceview_context_class_extents(self) -> None:
        manager = GtkSource.LanguageManager.get_default()
        language = manager.get_language("c")

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "int i = 0;\n// comment\nint j = 0;")
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(11)
        self.assertEqual(buffer.get_context_classes_at_iter(inside_iter), ["comment"])
        start, end = text_utils.get_iters_at_sourceview_context_class_extents(
            "comment", inside_iter
        )
        self.assertIsNotNone(start)
        assert start is not None  # mypy
        self.assertIsNotNone(end)
        assert end is not None  # mypy
        self.assertEqual(start.get_offset(), 11)
        self.assertEqual(end.get_offset(), 21)

    def test_parse_markdown_inline_link(self) -> None:
        _manager, language = self.__get_language()

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "plain")
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(3)
        self.assertIsNone(text_utils.parse_markdown_inline_link(inside_iter))

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "[text](https://linky.com)")
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(3)
        details = text_utils.parse_markdown_inline_link(inside_iter)
        self.assertIsNotNone(details)
        assert details is not None  # mypy
        self.assertEqual(details.link, "https://linky.com")
        self.assertEqual(details.text, "text")
        self.assertEqual(details.title, "")
        self.assertFalse(details.url_angle_bracketed)
        self.assertEqual(details.source_text, "[text](https://linky.com)")
        self.assertEqual(details.start_offset, 0)
        self.assertEqual(details.end_offset, 25)

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, '[text](https://linky.com "title")')
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(3)
        details = text_utils.parse_markdown_inline_link(inside_iter)
        self.assertIsNotNone(details)
        assert details is not None  # mypy
        self.assertEqual(details.link, "https://linky.com")
        self.assertEqual(details.text, "text")
        self.assertEqual(details.title, "title")
        self.assertFalse(details.url_angle_bracketed)
        self.assertEqual(details.source_text, '[text](https://linky.com "title")')
        self.assertEqual(details.start_offset, 0)
        self.assertEqual(details.end_offset, 33)

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "[text](<https://linky.com>)")
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(3)
        details = text_utils.parse_markdown_inline_link(inside_iter)
        self.assertIsNotNone(details)
        assert details is not None  # mypy
        self.assertEqual(details.link, "https://linky.com")
        self.assertEqual(details.text, "text")
        self.assertEqual(details.title, "")
        self.assertTrue(details.url_angle_bracketed)
        self.assertEqual(details.source_text, "[text](<https://linky.com>)")
        self.assertEqual(details.start_offset, 0)
        self.assertEqual(details.end_offset, 27)

    def test_parse_markdown_automatic_link(self) -> None:
        _manager, language = self.__get_language()

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "plain")
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(3)
        self.assertIsNone(text_utils.parse_markdown_automatic_link(inside_iter))

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, " <https://linky.com> ")
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(3)
        details = text_utils.parse_markdown_automatic_link(inside_iter)
        self.assertIsNotNone(details)
        assert details is not None  # mypy
        self.assertEqual(details.link, "https://linky.com")
        self.assertEqual(details.source_text, "<https://linky.com>")
        self.assertEqual(details.start_offset, 1)
        self.assertEqual(details.end_offset, 20)

        buffer = GtkSource.Buffer.new_with_language(language)
        start_iter = buffer.get_start_iter()
        self.__insert_into_buffer_and_wait(buffer, start_iter, "[text](<https://linky.com>)")
        inside_iter = buffer.get_start_iter()
        inside_iter.forward_chars(9)
        self.assertIsNone(text_utils.parse_markdown_automatic_link(inside_iter))

    def __insert_into_buffer_and_wait(
        self, buffer: GtkSource.Buffer, location: Gtk.TextIter, text: str
    ) -> None:
        context = GLib.MainContext.default()
        buffer.connect("highlight-updated", self.__flag_not_waiting)
        self.awaiting_signal = True
        buffer.insert(location, text)
        while self.awaiting_signal:
            context.iteration(False)

    def __flag_not_waiting(self, *_args) -> None:
        self.awaiting_signal = False

    def __get_language(self) -> tuple[GtkSource.LanguageManager, GtkSource.Language]:
        manager = GtkSource.LanguageManager()
        file_dir = os.path.dirname(__file__)
        language_path = os.path.join(
            file_dir, os.pardir, "data", "gtksourceview-5", "language-specs"
        )
        manager.prepend_search_path(language_path)
        language = manager.get_language("iotas-markdown")
        return (manager, language)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import os
import shutil
import time
from typing import Optional
import unittest
import warnings

from iotas.backup_manager import BackupManager
from iotas.backup_storage import BackupStorage
from iotas.note import Note, DirtyFields

warnings.filterwarnings("ignore", "version")


class _MockStorage(BackupStorage):

    notes: list[Note] = []

    def add_note(self, note: Note) -> None:
        """Add a new note to the database.

        :param Note note: Note to add
        """
        self.notes.append(note)

    def get_all_notes_count(self, include_locally_deleted: bool = True) -> int:
        """Fetch the number of notes in the database.

        :param bool include_locally_deleted: Whether to include locally deleted notes in the count
        :return: The count
        :rtype: int
        """
        return len(self.notes)

    def get_all_notes(self, load_content: bool = False) -> list[Note]:
        """Fetch all notes from the database.

        :param bool load_content: Whether to load the content for each note
        :return: List of notes
        :rtype: list[Note]
        """
        return self.notes[:]

    def get_note_by_remote_id(self, remote_id: int) -> Optional[Note]:
        """Fetch note from the database by remote id.

        :param int remote_id: Remote id of the note
        :return: The note, or None
        :rtype: Optional[Note]
        """
        result = None
        for note in self.notes:
            if note.remote_id == remote_id:
                result = note
                break
        return result

    def get_note_by_title(self, title: str) -> Optional[Note]:
        """Fetch note from the database by title.

        :param str title: Search title
        :return: The note, or None
        :rtype: Optional[Note]
        """
        result = None
        for note in self.notes:
            if note.title == title:
                result = note
                break
        return result

    def persist_note_selective(self, note: Note, updated_fields: DirtyFields) -> None:
        """Persist local note based on remote changes.

        :param Note note: The note to update
        :param DirtyFields updated_fields: Which fields to update
        """
        # Not implementing for unit test due to backup merging being disabled
        raise NotImplementedError()

    def create_duplicate_note(self, note: Note, reason: str) -> Note:
        """Create a duplicate note with a prefixed title.

        :param Note note: The note to duplicate
        :param str reason: The reason for the duplication, which is prefixed on the title
        """
        # Not implementing for unit test due to backup merging being disabled
        raise NotImplementedError()

    def flush(self) -> None:
        self.notes = []


class Test(unittest.TestCase):

    BACKUPS_SUBPATH = "backup"

    mock_storage = _MockStorage()

    def test_create_backup(self) -> None:
        self.__reset()
        backups_dir = os.path.join(self.__get_output_dir(), self.BACKUPS_SUBPATH)

        note1 = Note(new_note=True)
        note1.content = "content"
        note1.title = "Test Note Title"
        self.mock_storage.add_note(note1)

        note2 = Note(new_note=True)
        note2.content = "second content"
        note2.title = "Second Note Title"
        self.mock_storage.add_note(note2)

        result = self.manager.create_backup(sync_configured=False, file_extension="md")
        self.assertTrue(result)

        files = os.listdir(backups_dir)
        self.assertEqual(len(files), 4)
        self.assertTrue(f"{note1.title}.md" in files)
        self.assertTrue(f"{note1.title}.md.iota" in files)
        self.assertTrue(f"{note2.title}.md" in files)
        self.assertTrue(f"{note2.title}.md.iota" in files)

        with open(os.path.join(backups_dir, f"{note1.title}.md"), "r") as f:
            self.assertEqual(note1.content, f.read())

        with open(os.path.join(backups_dir, f"{note1.title}.md.iota"), "r") as f:
            meta_dict = json.loads(f.read())
        self.assertEqual(meta_dict["SchemaVersion"], "1.0")
        self.assertEqual(meta_dict["IotasVersion"], "0.0")
        self.assertEqual(meta_dict["Title"], note1.title)
        self.assertEqual(meta_dict["Category"], note1.category)
        self.assertEqual(meta_dict["LastModified"], note1.last_modified)
        self.assertEqual(meta_dict["Favourite"], note1.favourite)
        self.assertEqual(meta_dict["Dirty"], note1.dirty)
        self.assertEqual(meta_dict["LocallyDeleted"], note1.locally_deleted)
        self.assertEqual(meta_dict["RemoteId"], note1.remote_id)
        self.assertEqual(meta_dict["ETag"], note1.etag)

        self.__reset()

        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.category = "category"
        note.last_modified = int(time.time())
        note.favourite = True
        note.dirty = True
        note.locally_deleted = True
        note.remote_id = 13
        note.etag = "etag"
        self.mock_storage.add_note(note)

        result = self.manager.create_backup(sync_configured=False, file_extension="md")
        self.assertTrue(result)

        file_path = os.path.join(backups_dir, f"{note.title}.md.iota")
        self.assertTrue(os.path.exists(file_path))

        with open(file_path, "r") as f:
            meta_dict = json.loads(f.read())
        self.assertEqual(meta_dict["SchemaVersion"], "1.0")
        self.assertEqual(meta_dict["IotasVersion"], "0.0")
        self.assertEqual(meta_dict["Title"], note.title)
        self.assertEqual(meta_dict["Category"], note.category)
        self.assertEqual(meta_dict["LastModified"], note.last_modified)
        self.assertEqual(meta_dict["Favourite"], note.favourite)
        self.assertEqual(meta_dict["Dirty"], note.dirty)
        self.assertEqual(meta_dict["LocallyDeleted"], note.locally_deleted)
        self.assertEqual(meta_dict["RemoteId"], note.remote_id)
        self.assertEqual(meta_dict["ETag"], note.etag)

        self.__clean_output_dir()

    def test_restore_backup(self) -> None:
        self.__reset()
        backups_dir = os.path.join(self.__get_output_dir(), self.BACKUPS_SUBPATH)
        os.makedirs(backups_dir)
        with open(os.path.join(backups_dir, "Title.md"), "w") as f:
            f.write("content")
        with open(os.path.join(backups_dir, "Title.md.iota"), "w") as f:
            f.write("""{
                  "SchemaVersion": "1.0",
                  "IotasVersion": "0.0",
                  "Title": "Title",
                  "Category": "",
                  "LastModified": 0,
                  "Favourite": false,
                  "Dirty": false,
                  "LocallyDeleted": false,
                  "RemoteId": -1,
                  "ETag": ""
                }""")
        with open(os.path.join(backups_dir, "Second Title.md"), "w") as f:
            f.write("second content")
        with open(os.path.join(backups_dir, "Second Title.md.iota"), "w") as f:
            f.write("""{
                  "SchemaVersion": "1.0",
                  "IotasVersion": "0.0",
                  "Title": "Second Title",
                  "Category": "category",
                  "LastModified": 1745975049,
                  "Favourite": true,
                  "Dirty": true,
                  "LocallyDeleted": true,
                  "RemoteId": 13,
                  "ETag": "etag"
                }""")
        result = self.manager.restore_backup(sync_configured=False)
        self.assertTrue(result)
        self.assertEqual(self.mock_storage.get_all_notes_count(True), 2)

        note = self.mock_storage.get_note_by_title("Title")
        self.assertIsNotNone(note)
        assert note  # mypy
        self.assertEqual(note.title, "Title")
        self.assertEqual(note.content, "content")
        self.assertEqual(note.category, "")
        self.assertEqual(note.last_modified, 0)
        self.assertEqual(note.favourite, False)
        self.assertEqual(note.dirty, False)
        self.assertEqual(note.locally_deleted, False)
        self.assertEqual(note.remote_id, -1)
        self.assertEqual(note.etag, "")

        note = self.mock_storage.get_note_by_title("Second Title")
        self.assertIsNotNone(note)
        assert note  # mypy
        self.assertEqual(note.title, "Second Title")
        self.assertEqual(note.content, "second content")
        self.assertEqual(note.category, "category")
        self.assertEqual(note.last_modified, 1745975049)
        self.assertEqual(note.favourite, True)
        self.assertEqual(note.dirty, True)
        self.assertEqual(note.locally_deleted, True)

        # Not populated during restore
        self.assertEqual(note.remote_id, -1)
        self.assertEqual(note.etag, "")

        result = self.manager.restore_backup(sync_configured=False)
        self.assertFalse(result)

        self.__clean_output_dir()

    def __reset(self) -> BackupManager:
        self.mock_storage.flush()
        self.__prepare_output_dir()
        file_dir = os.path.dirname(__file__)
        storage_dir = os.path.join(file_dir, os.pardir, "testing-tmp")
        self.manager = BackupManager(
            self.mock_storage,
            iotas_version="0.0",
            storage_dir=storage_dir,
        )
        return self.manager

    def __prepare_output_dir(self) -> str:
        out_path = self.__get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)
        os.makedirs(out_path)
        return out_path

    def __clean_output_dir(self) -> None:
        out_path = self.__get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)

    def __get_output_dir(self) -> str:
        file_dir = os.path.dirname(__file__)
        return os.path.join(file_dir, os.pardir, "testing-tmp")



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, GLib, GObject, Gtk

import logging
import time


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/table_dialog.ui")
class TableDialog(Adw.Dialog):
    """Table creation dialog."""

    __gtype_name__ = "TableDialog"

    __gsignals__ = {
        "create": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
    }

    GRID_WIDTH = 6
    GRID_HEIGHT = 5
    DEFAULT_HEIGHT = 3
    DEFAULT_WIDTH = 3

    _size_grid: Gtk.Grid = Gtk.Template.Child()
    _size_label: Gtk.Label = Gtk.Template.Child()
    _button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()
        self.__width = self.DEFAULT_WIDTH
        self.__height = self.DEFAULT_HEIGHT
        self.__hover_timeout_id = None
        self.__hover_until = 0.0
        self.__build()
        self.connect("closed", lambda _o: self.__cleanup())

    def show(self, parent: Gtk.Window) -> None:
        """Show dialog.

        :param Gtk.Window window: Parent window
        """
        self.width = self.DEFAULT_WIDTH
        self.height = self.DEFAULT_HEIGHT
        for action in self.__action_group.list_actions():
            self.__action_group.lookup_action(action).set_enabled(True)
        self.present(parent)

    @GObject.Property(type=int)
    def width(self) -> int:
        return self.__width

    @width.setter
    def set_width(self, value: int) -> None:
        self.__width = value
        self.__update_grid()

    @GObject.Property(type=int)
    def height(self) -> int:
        return self.__height

    @height.setter
    def set_height(self, value: int) -> None:
        self.__height = value
        self.__update_grid()

    @Gtk.Template.Callback()
    def _on_button_clicked(self, _obj: Gtk.Button) -> None:
        self.__create()

    def __build(self) -> None:
        for row in range(self.GRID_HEIGHT):
            for column in range(self.GRID_WIDTH):
                button = Gtk.Button.new()
                self._size_grid.attach(button, column, row, 1, 1)

                def clicked(_o, width, height) -> None:
                    logging.info(f"Chose {width}x{height}")
                    self.width = width
                    self.height = height

                button.connect("clicked", clicked, column + 1, row + 1)

                controller = Gtk.EventControllerMotion.new()
                controller.connect("enter", self.__on_motion_enter, row, column)
                button.add_controller(controller)

        controller = Gtk.EventControllerMotion.new()
        controller.connect("leave", lambda _o: self.__on_motion_leave())
        self._size_grid.add_controller(controller)

        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("submit")
        action.connect("activate", lambda _a, _p: self.__create())
        action_group.add_action(action)
        app.set_accels_for_action("table-dialog.submit", ["<Alt>Return"])

        app.get_active_window().insert_action_group("table-dialog", action_group)
        self.__action_group = action_group

    def __update_grid(self) -> None:
        label = f"{self.width} x {self.height}"
        self._size_label.set_text(label)
        for row in range(self.GRID_HEIGHT):
            for column in range(self.GRID_WIDTH):
                button = self._size_grid.get_child_at(column, row)
                if self.width >= column + 1 and self.height >= row + 1:
                    button.add_css_class("on")
                else:
                    button.remove_css_class("on")

    def __on_motion_enter(
        self, motion: Gtk.EventControllerMotion, x: float, y: float, row: int, column: int
    ) -> None:
        # Hacky avoidance of motion on possibly mobile device. Without this (in Sept. 2024) a hover
        # pseudoclass is left on the buttons.
        if self.get_width() <= 360:
            return

        # If no motion hide hover after a second
        self.__hover_until = time.time() + 1
        if self.__hover_timeout_id is None:
            self.__hover_timeout_id = GLib.timeout_add(
                1000,
                self.__hover_timeout,
            )

        self.__update_hover(row, column)

    def __on_motion_leave(self) -> None:
        if self.__hover_timeout_id is not None:
            GLib.source_remove(self.__hover_timeout_id)
            self.__hover_timeout_id = None
        self.__update_hover(-1, -1)

    def __update_hover(self, hover_row: int, hover_column: int) -> None:
        for row in range(self.GRID_HEIGHT):
            for column in range(self.GRID_WIDTH):
                button = self._size_grid.get_child_at(column, row)
                if hover_column >= column and hover_row >= row:
                    button.add_css_class("hovering")
                else:
                    button.remove_css_class("hovering")

    def __hover_timeout(self) -> None:
        extra_hover = self.__hover_until - time.time()
        if extra_hover > 0.0:
            self.__hover_timeout_id = GLib.timeout_add(
                extra_hover * 1000.0,
                self.__hover_timeout,
            )
            return

        self.__update_hover(-1, -1)
        self.__hover_timeout_id = None

    def __create(self) -> None:
        self.emit("create", self.width, self.height)
        self.close()

    def __cleanup(self) -> None:
        for action in self.__action_group.list_actions():
            self.__action_group.lookup_action(action).set_enabled(False)
